@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * Authors:
  *   Chema Celorio <chema@celorio.com>
@@ -39,6 +39,12 @@ enum
 	LAST_SIGNAL
 };
 
+typedef enum {
+	CELL_ICON,
+	CELL_NAME,
+	CELL_MISC
+} GPVCellType;
+
 static GtkScrolledWindow *parent_class = NULL;
 
 static void glade_project_view_class_init (GladeProjectViewClass *class);
@@ -189,7 +195,10 @@ glade_project_view_populate_model (GladeProjectView *view)
 	for (list = project->objects; list; list = list->next)
 	{
 		GObject *object = G_OBJECT (list->data);
-		if (GTK_IS_WIDGET (object) && GTK_WIDGET_TOPLEVEL (object))
+		GladeWidget *gwidget = glade_widget_get_from_gobject (object);
+		g_assert (gwidget);
+
+		if (gwidget->parent == NULL)
 			toplevels = g_list_append (toplevels, object);
 	}
 
@@ -517,29 +526,43 @@ glade_project_view_cell_function (GtkTreeViewColumn *tree_column,
 				  GtkTreeIter *iter,
 				  gpointer data)
 {
-	gboolean     is_icon = GPOINTER_TO_INT (data);
+	GPVCellType  type = GPOINTER_TO_INT (data);
 	GladeWidget *widget;
+	gchar       *text = NULL, *child_type;
 
 	gtk_tree_model_get (tree_model, iter, WIDGET_COLUMN, &widget, -1);
 
 	/* The cell exists, but not widget has been asociated with it */
-	if (!widget)
-		return;
+	if (!widget) return;
 
 	g_return_if_fail (widget->name != NULL);
 	g_return_if_fail (widget->widget_class != NULL);
 	g_return_if_fail (widget->widget_class->name != NULL);
 	g_return_if_fail (widget->widget_class->icon != NULL);
 
-	if (is_icon)
-	{
-		g_object_set (G_OBJECT (cell), 
-			      "pixbuf", widget->widget_class->icon, 
-			      NULL);
-	}
-	else
+	switch (type) 
 	{
+	case CELL_ICON:
+		g_object_set (G_OBJECT (cell), "pixbuf", 
+			      widget->widget_class->icon, NULL);
+		break;
+	case CELL_NAME:
 		g_object_set (G_OBJECT (cell), "text", widget->name, NULL);
+		break;
+	case CELL_MISC:
+		/* special child type / internal child */
+		if (glade_widget_get_internal (widget) != NULL)
+			text = g_strdup_printf (_("(internal %s)"),  
+						glade_widget_get_internal (widget));
+		else if ((child_type = g_object_get_data (glade_widget_get_object (widget),
+							  "special-child-type")) != NULL)
+			text = g_strdup_printf (_("(%s child)"), child_type);
+
+		g_object_set (G_OBJECT (cell), "text", text ? text : " ", NULL);
+		if (text) g_free (text);
+		break;
+	default:
+		break;
 	}
 }
 
@@ -556,14 +579,22 @@ glade_project_view_add_columns (GtkTreeView *view)
 	gtk_tree_view_column_pack_start (column, renderer, FALSE);
 	gtk_tree_view_column_set_cell_data_func (column, renderer,
 						 glade_project_view_cell_function,
-						 GINT_TO_POINTER (1), NULL);
-	
+						 GINT_TO_POINTER (CELL_ICON), NULL);
+
 	renderer = gtk_cell_renderer_text_new ();
 	gtk_tree_view_column_pack_start (column, renderer, TRUE);
 	gtk_tree_view_column_set_cell_data_func (column, renderer,
 						 glade_project_view_cell_function,
-						 GINT_TO_POINTER (0), NULL);
+						 GINT_TO_POINTER (CELL_NAME), NULL);
 
+	renderer = gtk_cell_renderer_text_new ();
+	g_object_set (G_OBJECT (renderer), 
+		      "style", PANGO_STYLE_ITALIC,
+		      "foreground", "Gray", NULL);
+	gtk_tree_view_column_pack_start (column, renderer, TRUE);
+	gtk_tree_view_column_set_cell_data_func (column, renderer,
+						 glade_project_view_cell_function,
+						 GINT_TO_POINTER (CELL_MISC), NULL);
 	gtk_tree_view_append_column (view, column);
 }
 
@@ -587,7 +618,6 @@ glade_project_view_init (GladeProjectView *view)
 	
 	view->selected_widget = NULL;
 	view->project = NULL;
-	view->add_widget_signal_id = 0;
 
 	view->model = gtk_tree_store_new (N_COLUMNS, G_TYPE_POINTER);
 	glade_project_view_populate_model (view);
@@ -670,6 +700,16 @@ glade_project_view_selection_update_cb (GladeProject *project,
 	GLADE_PROJECT_VIEW_GET_CLASS (view)->selection_update (view, project);
 }
 
+static void
+glade_project_view_close_cb (GladeProject *project,
+			     GladeProjectView *view)
+{
+	/* Detected project we are viewing was closed,
+	 * detatch from view.
+	 */
+	glade_project_view_set_project (view, NULL);
+}
+
 /**
  * glade_project_view_new:
  * @type: a #GladeProjectViewType
@@ -729,6 +769,8 @@ glade_project_view_set_project (GladeProjectView *view,
 					     view->widget_name_changed_signal_id);
 		g_signal_handler_disconnect (G_OBJECT (view->project),
 					     view->selection_changed_signal_id);
+		g_signal_handler_disconnect (G_OBJECT (view->project),
+					     view->project_closed_signal_id);
 	}
 
 	model = GTK_TREE_MODEL (view->model);
@@ -770,6 +812,10 @@ glade_project_view_set_project (GladeProjectView *view,
 		g_signal_connect (G_OBJECT (project), "selection_changed",
 				  G_CALLBACK (glade_project_view_selection_update_cb),
 				  view);
+	view->project_closed_signal_id =
+		g_signal_connect (G_OBJECT (project), "close",
+				  G_CALLBACK (glade_project_view_close_cb),
+				  view);
 }
 
 /**
