@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * Authors:
  *   Chema Celorio <chema@celorio.com>
@@ -80,11 +80,15 @@ static void
 gpw_refresh_title (GladeProjectWindow *gpw)
 {
 	GladeProject *active_project;
-	gchar *title;
+	gchar *title, *disp;
 
 	active_project = glade_app_get_active_project (GLADE_APP (gpw));
 	if (active_project)
-		title = g_strdup_printf ("%s - %s", WINDOW_TITLE, active_project->name);
+	{
+		disp = glade_project_display_name (active_project);
+		title = g_strdup_printf ("%s - %s", WINDOW_TITLE, disp);
+		g_free (disp);
+	}
 	else
 		title = g_strdup_printf ("%s", WINDOW_TITLE);
 
@@ -92,6 +96,26 @@ gpw_refresh_title (GladeProjectWindow *gpw)
 	g_free (title);
 }
 
+static GtkWidget *
+gpw_menuitem_from_project (GladeProject *project)
+{
+	GtkUIManager *ui;
+	gchar        *path;
+
+	g_return_val_if_fail (GLADE_IS_PROJECT (project), NULL);
+
+	ui   = g_object_get_data (G_OBJECT (project), "ui");
+	path = g_object_get_data (G_OBJECT (project), "menuitem_path");
+	return gtk_ui_manager_get_widget (ui, path);
+}
+
+static guint
+gpw_menuitem_merge_id_from_project (GladeProject *project)
+{
+	g_return_val_if_fail (GLADE_IS_PROJECT (project), 0);
+	return GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (project), "merge_id"));
+}
+
 static void
 gpw_select_project_menu (GladeProjectWindow *gpw)
 {
@@ -103,7 +127,7 @@ gpw_select_project_menu (GladeProjectWindow *gpw)
 	     projects && projects->data; projects = projects->next)
 	{
 		project      = projects->data;
-		palette_item = glade_project_get_menuitem (project);
+		palette_item = gpw_menuitem_from_project (project);
 		
 		gtk_check_menu_item_set_draw_as_radio
 			(GTK_CHECK_MENU_ITEM (palette_item), TRUE);
@@ -117,16 +141,16 @@ gpw_select_project_menu (GladeProjectWindow *gpw)
 static void
 gpw_refresh_project_entry (GladeProjectWindow *gpw, GladeProject *project)
 {
+	GtkAction *action = g_object_get_data (G_OBJECT (project), "action");
+
 	/* Remove menuitem and action */
 	gtk_ui_manager_remove_ui(gpw->priv->ui,
-				 glade_project_get_menuitem_merge_id(project));
+				 gpw_menuitem_merge_id_from_project (project));
 
-	gtk_action_group_remove_action (gpw->priv->project_actions,
-					GTK_ACTION (project->action));
-	
-	g_object_unref (G_OBJECT (project->action));
+	gtk_action_group_remove_action (gpw->priv->project_actions, action);
+	g_object_unref (G_OBJECT (action));
 
-	/* Create project menuiten and action */
+	/* Create project menuitem and action */
 	gpw_project_menuitem_add (gpw, project);
 	
 	gpw_select_project_menu (gpw);
@@ -144,11 +168,8 @@ gpw_recent_project_delete (GtkAction *action, GladeProjectWindow *gpw)
 	guint merge_id = GPOINTER_TO_UINT(g_object_get_data (G_OBJECT (action), "merge_id"));
 	
 	gtk_ui_manager_remove_ui(gpw->priv->ui,	merge_id);
-
 	gtk_action_group_remove_action (gpw->priv->recent_actions, action);
-	
 	g_queue_remove (gpw->priv->recent_projects, action);
-	
 	gtk_ui_manager_ensure_update (gpw->priv->ui);
 }
 
@@ -295,22 +316,8 @@ gpw_open_cb (GtkAction *action, GladeProjectWindow *gpw)
 
 	if (!path)
 		return;
-
-	/* dont allow more than one project with the same name to be
-	 * opened simultainiously.
-	 */
-	if (glade_app_is_project_loaded (GLADE_APP (gpw), path))
-	{
-		gchar *base_path = g_path_get_basename (path);
-		gchar *message = g_strdup_printf (_("A project named %s is already open"), base_path);
-
-		glade_util_ui_warn (gpw->priv->window, message);
-
-		g_free (message);
-		g_free (base_path);
-	} else {
-		glade_project_window_open_project (gpw, path);
-	}
+	
+	glade_project_window_open_project (gpw, path);
 	g_free (path);
 }
 
@@ -351,16 +358,28 @@ gpw_recent_project_clear_cb (GtkAction *action, GladeProjectWindow *gpw)
 static void
 gpw_save (GladeProjectWindow *gpw, GladeProject *project, const gchar *path)
 {
-	GError              *error = NULL;
-	
+	GError  *error = NULL;
+	gchar   *display_name = glade_project_display_name (project),
+		*display_path = g_strdup (path);
+
+	/* Interestingly; we cannot use `path' after glade_project_reset_path
+	 * because we are getting called with project->path as an argument.
+	 */
 	if (!glade_project_save (project, path, &error))
 	{
-		glade_util_ui_warn (gpw->priv->window, error->message);
+		/* Reset path so future saves will prompt the file chooser */
+		glade_project_reset_path (project);
+		glade_util_ui_message (gpw->priv->window, GLADE_UI_ERROR, 
+				       _("Failed to save %s: %s"),
+				       display_path, error->message);
 		g_error_free (error);
+		g_free (display_path);
 		return;
 	}
 	
-	gpw_recent_project_add (gpw, path);
+	glade_app_update_instance_count (GLADE_APP (gpw), project);
+
+	gpw_recent_project_add (gpw, project->path);
 	gpw_recent_project_config_save (gpw);
 	glade_app_config_save (GLADE_APP (gpw));
 
@@ -368,7 +387,10 @@ gpw_save (GladeProjectWindow *gpw, GladeProject *project, const gchar *path)
 	gpw_refresh_title (gpw);
 	glade_util_flash_message (gpw->priv->statusbar,
 				  gpw->priv->statusbar_actions_context_id,
-				  _("Project '%s' saved"), project->name);
+				  _("Project '%s' saved"), display_name);
+
+	g_free (display_path);
+	g_free (display_name);
 }
 
 static void
@@ -380,8 +402,13 @@ gpw_save_as (GladeProjectWindow *gpw, const gchar *dialog_title)
 	gchar *real_path;
 	gchar *ch;
 	
-	project = glade_app_get_active_project (GLADE_APP (gpw));
-	
+	if ((project = glade_app_get_active_project (GLADE_APP (gpw))) == NULL)
+	{
+		glade_util_ui_message (gpw->priv->window, GLADE_UI_WARN, 
+				       _("No open projects to save"));
+		return;
+	}
+
 	filechooser = glade_util_file_dialog_new (dialog_title,
 						  GTK_WINDOW (gpw->priv->window),
 						  GLADE_FILE_DIALOG_ACTION_SAVE);
@@ -402,50 +429,34 @@ gpw_save_as (GladeProjectWindow *gpw, const gchar *dialog_title)
 	
 	g_free (path);
 
-	/* dont allow more than one project with the same name to be
-	 * opened simultainiously (but allow the same project to change paths).
-	 */
-	if (glade_app_is_project_loaded (GLADE_APP (gpw), real_path) &&
-	    glade_util_basenames_match  (project->path, real_path) == FALSE)
+	if (glade_app_is_project_loaded (GLADE_APP (gpw), real_path))
 	{
-		gchar *base_path = g_path_get_basename (path);
-		gchar *message = g_strdup_printf (_("A project named %s is already open"), base_path);
-
-		glade_util_ui_warn (gpw->priv->window, message);
-
-		g_free (message);
-		g_free (base_path);
+		glade_util_ui_message (gpw->priv->window, GLADE_UI_WARN,
+				       _("%s is already open"), real_path);
 		return;
 	}
-
 	gpw_save (gpw, project, real_path);
 
 	g_free (real_path);
 }
 
+
 static void
 gpw_save_cb (GtkAction *action, GladeProjectWindow *gpw)
 {
 	GladeProject       *project;
-	GError             *error = NULL;
-	
-	project = glade_app_get_active_project (GLADE_APP (gpw));
+
+	if ((project = glade_app_get_active_project (GLADE_APP (gpw))) == NULL)
+	{
+		glade_util_ui_message (gpw->priv->window, 
+				       GLADE_UI_WARN,
+				       _("No open projects to save"));
+		return;
+	}
 
 	if (project->path != NULL) 
 	{
-		if (glade_project_save (project, project->path, &error)) 
-		{
-			glade_util_flash_message 
-				(gpw->priv->statusbar,
-				 gpw->priv->statusbar_actions_context_id,
-				 _("Project '%s' saved"),
-				 project->name);
-		} 
-		else 
-		{
-			glade_util_ui_warn (gpw->priv->window, error->message);
-			g_error_free (error);
-		}
+		gpw_save (gpw, project, project->path);
  		return;
 	}
 
@@ -502,7 +513,11 @@ gpw_confirm_close_project (GladeProjectWindow *gpw, GladeProject *project)
 			if ((close = glade_project_save
 			     (project, project->path, &error)) == FALSE)
 			{
-				glade_util_ui_warn (gpw->priv->window, error->message);
+
+				glade_util_ui_message
+					(gpw->priv->window, GLADE_UI_ERROR, 
+					 _("Failed to save %s to %s: %s"),
+					 project->name, project->path, error->message);
 				g_error_free (error);
 			}
 		}
@@ -549,7 +564,7 @@ static void
 do_close (GladeProjectWindow *gpw, GladeProject *project)
 {
 	gtk_ui_manager_remove_ui (gpw->priv->ui,
-				  glade_project_get_menuitem_merge_id (project));
+				  gpw_menuitem_merge_id_from_project (project));
 	
 	glade_app_remove_project (GLADE_APP (gpw), project);
 	gpw_select_project_menu (gpw);
@@ -676,7 +691,6 @@ gpw_create_palette (GladeProjectWindow *gpw)
 	gtk_window_set_default_size (GTK_WINDOW (gpw->priv->palette_window), -1,
 				     GLADE_PALETTE_DEFAULT_HEIGHT);
 
-	gtk_window_set_transient_for (gpw->priv->palette_window, GTK_WINDOW (gpw->priv->window));
 
 	gtk_window_set_title (gpw->priv->palette_window, _("Palette"));
 	gtk_window_set_type_hint (gpw->priv->palette_window, GDK_WINDOW_TYPE_HINT_UTILITY);
@@ -738,6 +752,21 @@ gpw_hide_editor_on_delete (GtkWidget *editor, gpointer not_used,
 	return TRUE;
 }
 
+static gint
+gpw_hijack_editor_key_press (GtkWidget          *editor_win, 
+			     GdkEventKey        *event, 
+			     GladeProjectWindow *gpw)
+{
+	/* If we recieve GDK_Delete, send it to the focus widget
+	 * and return TRUE so that the accel_group doesn't kick in.
+	 */
+	if (event->keyval == GDK_Delete)
+	{
+		return gtk_widget_event (GTK_WINDOW (editor_win)->focus_widget, (GdkEvent *)event);
+	}
+	return FALSE;
+}
+
 static void
 gpw_create_editor (GladeProjectWindow *gpw)
 {
@@ -748,8 +777,6 @@ gpw_create_editor (GladeProjectWindow *gpw)
 	gpw->priv->editor_window = GTK_WINDOW (gtk_window_new (GTK_WINDOW_TOPLEVEL));
 	gtk_window_set_default_size (GTK_WINDOW (gpw->priv->editor_window), 400, 450);
 
-	gtk_window_set_transient_for (gpw->priv->editor_window, GTK_WINDOW (gpw->priv->window));
-
 	gtk_window_set_title  (gpw->priv->editor_window, _("Properties"));
 	gtk_window_move (gpw->priv->editor_window, 350, 0);
 
@@ -760,6 +787,11 @@ gpw_create_editor (GladeProjectWindow *gpw)
 	g_signal_connect (G_OBJECT (gpw->priv->editor_window), "delete_event",
 			  G_CALLBACK (gpw_hide_editor_on_delete), gpw->priv->ui);
 
+
+	g_signal_connect (G_OBJECT (gpw->priv->editor_window), "key-press-event",
+			  G_CALLBACK (gpw_hijack_editor_key_press), gpw);
+
+
 	editor_item = gtk_ui_manager_get_widget (gpw->priv->ui,
 						 "/MenuBar/ViewMenu/PropertyEditor");
 	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (editor_item), TRUE);
@@ -876,8 +908,6 @@ gpw_create_clipboard_view (GladeProjectWindow *gpw)
 	
 	view = glade_app_get_clipboard_view (GLADE_APP (gpw));
 
-	gtk_window_set_transient_for (GTK_WINDOW (view), GTK_WINDOW (gpw->priv->window));
-
 	g_signal_connect (G_OBJECT (view), "delete_event",
 			  G_CALLBACK (gpw_hide_clipboard_view_on_delete),
 			  gpw->priv->ui);
@@ -1000,7 +1030,8 @@ static void gpw_about_cb (GtkAction *action, GladeProjectWindow *gpw)
 		  "GNU General Public License for more details.\n\n"
 		  "You should have received a copy of the GNU General Public License\n"
 		  "along with this program; if not, write to the Free Software\n"
-		  "Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.");
+		  "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, "
+		  "MA 02110-1301, USA.");
 
 	gchar *copyright =
 		"(C) 2001-2005 Ximian, Inc.\n"
@@ -1024,16 +1055,17 @@ glade_project_window_set_project (GtkAction *action, GladeProject *project)
 	GtkWidget *item;
 
 	gpw  = g_object_get_data (G_OBJECT (project), "gpw");	
-	item = glade_project_get_menuitem (project);
+	item = gpw_menuitem_from_project (project);
 
 	if (gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (item)))
 	{
 		glade_app_set_project (GLADE_APP (gpw), project);
 
-		/* Ensure correct selection
-		 */
-		gpw_select_project_menu (gpw);
 	}
+
+	/* Ensure correct selection
+	 */
+	gpw_select_project_menu (gpw);
 }
 
 static const gchar *ui_info =
@@ -1133,7 +1165,7 @@ static GtkActionEntry entries[] = {
 	"Delete the selection", G_CALLBACK (gpw_delete_cb) },
 	
 	/* HelpMenu */
-	{ "About", NULL, "_About", NULL,
+	{ "About", GTK_STOCK_ABOUT, "_About", NULL,
 	"Shows the About Dialog", G_CALLBACK (gpw_about_cb) }
 };
 static guint n_entries = G_N_ELEMENTS (entries);
@@ -1319,8 +1351,6 @@ glade_project_window_create (GladeProjectWindow *gpw)
 	g_free (filename);
 	gpw->priv->window = app;
 
-	glade_app_set_transient_parent (GLADE_APP (gpw), GTK_WINDOW (app));
-
 	vbox = gtk_vbox_new (FALSE, 0);
 	gtk_container_add (GTK_CONTAINER (app), vbox);
 	gpw->priv->main_vbox = vbox;
@@ -1358,45 +1388,47 @@ glade_project_window_create (GladeProjectWindow *gpw)
 static void
 gpw_project_menuitem_add (GladeProjectWindow *gpw, GladeProject *project)
 {
-	gchar *underscored_name, *action_name;
+	gchar *final_name, *action_name, *display_name;
+	GtkWidget *action;
 	guint merge_id;
 	
 	/* double the underscores in the project name
 	 * (or they will just underscore the next character) */
-	underscored_name = glade_util_duplicate_underscores (project->name);
-	if (!underscored_name)
+	display_name = glade_project_display_name (project);
+	if ((final_name = 
+	     glade_util_duplicate_underscores (display_name)) == NULL)
 		return;
 
-	action_name = g_strdup_printf ("select[%s]", project->name);
+	action_name = g_strdup_printf ("select[%s]", final_name);
 	
-	/* What happen if there is two project with the same name in differents
-	 * directories?
-	 */
-	project->action = gtk_toggle_action_new (action_name, underscored_name,
-					"Selects this Project",	NULL);
-	gtk_toggle_action_set_active (project->action, TRUE);
-	g_signal_connect (G_OBJECT (project->action), "activate",
+	action = (GtkWidget *)gtk_toggle_action_new (action_name, final_name,
+						     project->path, NULL);
+	g_object_set_data (G_OBJECT (project), "action", action);
+
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
+	g_signal_connect (G_OBJECT (action), "activate",
 			  (GCallback) glade_project_window_set_project, project);
 	
 	gtk_action_group_add_action_with_accel (gpw->priv->project_actions,
-						GTK_ACTION (project->action), "");
+						GTK_ACTION (action), "");
 	
 	/* Add menuitem to menu */
-	merge_id = gtk_ui_manager_new_merge_id(gpw->priv->ui);
+	merge_id = gtk_ui_manager_new_merge_id (gpw->priv->ui);
 	
 	gtk_ui_manager_add_ui(gpw->priv->ui, merge_id,
 			      "/MenuBar/ProjectMenu", action_name, action_name,
 			      GTK_UI_MANAGER_MENUITEM, FALSE);
 
 	/* Set extra data to action */
-	g_object_set_data (G_OBJECT (project->action), "ui", gpw->priv->ui);
-	g_object_set_data_full (G_OBJECT (project->action), "menuitem_path",
+	g_object_set_data (G_OBJECT (project), "ui", gpw->priv->ui);
+	g_object_set_data_full (G_OBJECT (project), "menuitem_path",
 				g_strdup_printf ("/MenuBar/ProjectMenu/%s", action_name),
 				g_free);
-	g_object_set_data (G_OBJECT (project->action), "merge_id", GINT_TO_POINTER (merge_id));
+	g_object_set_data (G_OBJECT (project), "merge_id", GINT_TO_POINTER (merge_id));
 
 	g_free (action_name);
-	g_free (underscored_name);
+	g_free (display_name);
+	g_free (final_name);
 }
 
 static void
@@ -1428,7 +1460,9 @@ glade_project_window_new_project (GladeProjectWindow *gpw)
 	project = glade_project_new (TRUE);
 	if (!project)
 	{
-		glade_util_ui_warn (gpw->priv->window, _("Could not create a new project."));
+		glade_util_ui_message (gpw->priv->window, 
+				       GLADE_UI_ERROR,
+				       _("Could not create a new project."));
 		return;
 	}
 	glade_project_window_add_project (gpw, project);
@@ -1451,16 +1485,16 @@ glade_project_window_open_project (GladeProjectWindow *gpw, const gchar *path)
 	 * opened simultainiously.
 	 */
 	if (glade_app_is_project_loaded (GLADE_APP (gpw), (gchar*)path))
-		return;
-
-	project = glade_project_open (path);
-	if (!project)
 	{
-		glade_util_ui_warn (gpw->priv->window, _("Could not open project."));
+		glade_util_ui_message (gpw->priv->window, GLADE_UI_WARN, 
+				       _("%s is already open"), path);
 		return;
 	}
+	
+	if ((project = glade_project_open (path)) == NULL)
+		return;
 
-	gpw_recent_project_add (gpw, path);
+	gpw_recent_project_add (gpw, project->path);
 	gpw_recent_project_config_save (gpw);
 	glade_app_config_save (GLADE_APP (gpw));
 
@@ -1518,41 +1552,50 @@ glade_project_window_change_menu_label (GladeProjectWindow *gpw,
 void
 glade_project_window_refresh_undo_redo (GladeProjectWindow *gpw)
 {
-	GList *prev_redo_item;
-	GList *undo_item;
-	GList *redo_item;
-	const gchar *undo_description = NULL;
-	const gchar *redo_description = NULL;
+	GladeCommand *undo = NULL, *redo = NULL;
 	GladeProject *project;
+	GtkAction    *action;
+	gchar        *desc;
 
-	project = glade_app_get_active_project (GLADE_APP (gpw));
-	if (!project)
+	if ((project = glade_app_get_active_project (GLADE_APP (gpw))) != NULL)
 	{
-		undo_item = NULL;
-		redo_item = NULL;
-	}
-	else
-	{
-		undo_item = prev_redo_item = project->prev_redo_item;
-		redo_item = (prev_redo_item == NULL) ? project->undo_stack : prev_redo_item->next;
-
-		if (undo_item && undo_item->data)
-			undo_description = GLADE_COMMAND (undo_item->data)->description;
-		if (redo_item && redo_item->data)
-			redo_description = GLADE_COMMAND (redo_item->data)->description;
+		undo = glade_command_next_undo_item (project);
+		redo = glade_command_next_redo_item (project);
 	}
 
-	glade_project_window_change_menu_label (gpw, "/MenuBar/EditMenu/Undo", _("_Undo: "), undo_description);
-	glade_project_window_change_menu_label (gpw, "/MenuBar/EditMenu/Redo", _("_Redo: "), redo_description);
+	/* Change label in menu */
+	glade_project_window_change_menu_label
+		(gpw, "/MenuBar/EditMenu/Undo", _("_Undo: "), undo ? undo->description : " ");
+
+	glade_project_window_change_menu_label
+		(gpw, "/MenuBar/EditMenu/Redo", _("_Redo: "), redo ? redo->description : " ");
+
+	
+	/* Change tooltips on the toolbar */
+	desc = g_strdup_printf (_("Undo: %s"), undo ? undo->description : _("the last action"));
+	action = gtk_ui_manager_get_action (gpw->priv->ui, "/ToolBar/Undo");
+	g_object_set (G_OBJECT (action), "tooltip", desc, NULL);
+	g_free (desc);
+
+	desc = g_strdup_printf (_("Redo: %s"), redo ? redo->description : _("the last action"));
+	action = gtk_ui_manager_get_action (gpw->priv->ui, "/ToolBar/Redo");
+	g_object_set (G_OBJECT (action), "tooltip", desc, NULL);
+	g_free (desc);
+
+	/* Set sensitivity on the toolbar */
+	gtk_widget_set_sensitive (gpw->priv->toolbar_undo, undo != NULL);
+	gtk_widget_set_sensitive (gpw->priv->toolbar_redo, redo != NULL);
 
-	gtk_widget_set_sensitive (gpw->priv->toolbar_undo, undo_description != NULL);
-	gtk_widget_set_sensitive (gpw->priv->toolbar_redo, redo_description != NULL);
 }
 
 static void
 glade_project_window_update_ui (GladeApp *app)
 {
 	GladeProjectWindow *gpw = GLADE_PROJECT_WINDOW (app);
+	
+	/* Chain Up */
+	GLADE_APP_CLASS (parent_class)->update_ui_signal (app);
+	
 	gpw_refresh_title (gpw);
 	glade_project_window_refresh_undo_redo (gpw);
 }
