@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * Authors:
  *   Chema Celorio <chema@celorio.com>
@@ -26,6 +26,9 @@
 
 #include <string.h>
 #include <stdlib.h>
+#include <glib.h>
+#include <glib/gi18n-lib.h>
+#include <glib/gstdio.h>
 
 #include "glade.h"
 #include "glade-project.h"
@@ -38,6 +41,7 @@
 #include "glade-utils.h"
 #include "glade-id-allocator.h"
 #include "glade-app.h"
+#include "glade-catalog.h"
 
 static void glade_project_class_init (GladeProjectClass *class);
 static void glade_project_init (GladeProject *project);
@@ -50,6 +54,9 @@ enum
 	REMOVE_WIDGET,
 	WIDGET_NAME_CHANGED,
 	SELECTION_CHANGED,
+	CLOSE,
+	RESOURCE_UPDATED,
+	RESOURCE_REMOVED,
 	LAST_SIGNAL
 };
 
@@ -140,6 +147,39 @@ glade_project_class_init (GladeProjectClass *class)
 			      G_TYPE_NONE,
 			      0);
 
+	glade_project_signals[CLOSE] =
+		g_signal_new ("close",
+			      G_TYPE_FROM_CLASS (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (GladeProjectClass, close),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__VOID,
+			      G_TYPE_NONE,
+			      0);
+
+	glade_project_signals[RESOURCE_UPDATED] =
+		g_signal_new ("resource-updated",
+			      G_TYPE_FROM_CLASS (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (GladeProjectClass, resource_updated),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__STRING,
+			      G_TYPE_NONE,
+			      1,
+			      G_TYPE_STRING);
+
+	glade_project_signals[RESOURCE_REMOVED] =
+		g_signal_new ("resource-removed",
+			      G_TYPE_FROM_CLASS (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (GladeProjectClass, resource_removed),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__STRING,
+			      G_TYPE_NONE,
+			      1,
+			      G_TYPE_STRING);
+
+
 	object_class->finalize = glade_project_finalize;
 	object_class->dispose = glade_project_dispose;
 	
@@ -147,19 +187,32 @@ glade_project_class_init (GladeProjectClass *class)
 	class->remove_object = NULL;
 	class->widget_name_changed = NULL;
 	class->selection_changed = NULL;
+	class->close = NULL;
+	class->resource_updated = NULL;
+	class->resource_removed = NULL;
 }
 
 static void
 glade_project_init (GladeProject *project)
 {
+	project->path = NULL;
 	project->name = NULL;
+	project->instance = 0;
 	project->objects = NULL;
 	project->selection = NULL;
 	project->undo_stack = NULL;
 	project->prev_redo_item = NULL;
-	project->widget_names_allocator = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) glade_id_allocator_free);
-	project->widget_old_names = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify) g_free);
+	project->widget_names_allocator = 
+		g_hash_table_new_full (g_str_hash, g_str_equal, g_free, 
+				       (GDestroyNotify) glade_id_allocator_free);
+	project->widget_old_names = 
+		g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify) g_free);
 	project->tooltips = gtk_tooltips_new ();
+	project->accel_group = NULL;
+
+	project->resources = g_hash_table_new_full (g_direct_hash, 
+						    g_direct_equal, 
+						    NULL, g_free);
 }
 
 /**
@@ -203,7 +256,10 @@ glade_project_dispose (GObject *object)
 	GladeProject *project = GLADE_PROJECT (object);
 	GList        *list;
 	GladeWidget  *gwidget;
-
+	
+	/* Emit close signal */
+	g_signal_emit (object, glade_project_signals [CLOSE], 0);
+	
 	glade_project_selection_clear (project, TRUE);
 
 	glade_project_list_unref (project->undo_stack);
@@ -235,6 +291,7 @@ glade_project_finalize (GObject *object)
 
 	g_hash_table_destroy (project->widget_names_allocator);
 	g_hash_table_destroy (project->widget_old_names);
+	g_hash_table_destroy (project->resources);
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -276,6 +333,81 @@ glade_project_on_widget_notify (GladeWidget *widget, GParamSpec *arg, GladeProje
 	}
 }
 
+
+static void
+gp_sync_resources (GladeProject *project, 
+		   GladeProject *prev_project,
+		   GladeWidget  *gwidget,
+		   gboolean      remove)
+{
+	GList          *prop_list, *l;
+	GladeProperty  *property;
+	gchar          *resource, *full_resource;
+
+	prop_list = g_list_copy (gwidget->properties);
+	prop_list = g_list_concat
+		(prop_list, g_list_copy (gwidget->packing_properties));
+
+	for (l = prop_list; l; l = l->next)
+	{
+		property = l->data;
+		if (property->class->resource)
+		{
+			GValue value = { 0, };
+
+			if (remove)
+			{
+				glade_project_set_resource (project, property, NULL);
+				continue;
+			}
+
+			glade_property_get_value (property, &value);
+			
+			if ((resource = glade_property_class_make_string_from_gvalue
+			     (property->class, &value)) != NULL)
+			{
+				full_resource = glade_project_resource_fullpath
+					(prev_project ? prev_project : project, resource);
+			
+				/* Use a full path here so that the current
+				 * working directory isnt used.
+				 */
+				glade_project_set_resource (project, 
+							    property,
+							    full_resource);
+				
+				g_free (full_resource);
+				g_free (resource);
+			}
+			g_value_unset (&value);
+		}
+	}
+	g_list_free (prop_list);
+}
+
+static void
+glade_project_sync_resources_for_widget (GladeProject *project, 
+					 GladeProject *prev_project,
+					 GladeWidget  *gwidget,
+					 gboolean      remove)
+{
+	GList *children, *l;
+	GladeWidget *gchild;
+
+	children = glade_widget_class_container_get_all_children
+		(gwidget->widget_class, gwidget->object);
+
+	for (l = children; l; l = l->next)
+		if ((gchild = 
+		     glade_widget_get_from_gobject (l->data)) != NULL)
+			glade_project_sync_resources_for_widget 
+				(project, prev_project, gchild, remove);
+	if (children)
+		g_list_free (children);
+
+	gp_sync_resources (project, prev_project, gwidget, remove);
+}
+
 /**
  * glade_project_add_object:
  * @project: the #GladeProject the widget is added to
@@ -284,17 +416,20 @@ glade_project_on_widget_notify (GladeWidget *widget, GParamSpec *arg, GladeProje
  * Adds an object to the project.
  */
 void
-glade_project_add_object (GladeProject *project, GObject *object)
+glade_project_add_object (GladeProject *project, 
+			  GladeProject *old_project, 
+			  GObject      *object)
 {
-	GladeWidget          *gwidget;
-	GList                *list, *children;
-	GtkWindow            *transient_parent;
+	GladeWidget   *gwidget;
+	GList         *list, *children;
+	GtkWindow     *transient_parent;
+	static gint    reentrancy_count = 0;
 
 	g_return_if_fail (GLADE_IS_PROJECT (project));
 	g_return_if_fail (G_IS_OBJECT      (object));
 
 	/* We don't list placeholders */
-	if (GLADE_IS_PLACEHOLDER (object))
+	if (GLADE_IS_PLACEHOLDER (object)) 
 		return;
 
 	/* Only widgets accounted for in the catalog or widgets declared
@@ -304,11 +439,15 @@ glade_project_add_object (GladeProject *project, GObject *object)
 	if ((gwidget = glade_widget_get_from_gobject (object)) == NULL)
 		return;
 
+	/* Code body starts here */
+	reentrancy_count++;
+
 	if ((children = glade_widget_class_container_get_all_children
 	     (gwidget->widget_class, gwidget->object)) != NULL)
 	{
 		for (list = children; list && list->data; list = list->next)
-			glade_project_add_object (project, G_OBJECT (list->data));
+			glade_project_add_object
+				(project, old_project, G_OBJECT (list->data));
 		g_list_free (children);
 	}
 
@@ -329,6 +468,14 @@ glade_project_add_object (GladeProject *project, GObject *object)
 	if (GTK_IS_WINDOW (object) &&
 	    (transient_parent = glade_default_app_get_transient_parent ()) != NULL)
 		gtk_window_set_transient_for (GTK_WINDOW (object), transient_parent);
+
+	/* Notify widget it was added to the project */
+	glade_widget_project_notify (gwidget, project);
+
+	/* Call this once at the end for every recursive call */
+	if (--reentrancy_count == 0)
+		glade_project_sync_resources_for_widget
+			(project, old_project, gwidget, FALSE);
 }
 
 /**
@@ -413,8 +560,9 @@ glade_project_release_widget_name (GladeProject *project, GladeWidget *glade_wid
 void
 glade_project_remove_object (GladeProject *project, GObject *object)
 {
-	GladeWidget          *gwidget;
-	GList                *link, *list, *children;
+	GladeWidget   *gwidget;
+	GList         *link, *list, *children;
+	static gint    reentrancy_count = 0;
 	
 	g_return_if_fail (GLADE_IS_PROJECT (project));
 	g_return_if_fail (G_IS_OBJECT      (object));
@@ -424,6 +572,12 @@ glade_project_remove_object (GladeProject *project, GObject *object)
 
 	if ((gwidget = glade_widget_get_from_gobject (object)) == NULL)
 		return;
+
+	/* Code body starts here */
+	reentrancy_count++;
+
+	/* Notify widget is being removed from the project */
+	glade_widget_project_notify (gwidget, NULL);
 	
 	if ((children = glade_widget_class_container_get_children (gwidget->widget_class,
 								   gwidget->object)) != NULL)
@@ -448,6 +602,10 @@ glade_project_remove_object (GladeProject *project, GObject *object)
 		       glade_project_signals [REMOVE_WIDGET],
 		       0,
 		       gwidget);
+
+	/* Call this once at the end for every recursive call */
+	if (--reentrancy_count == 0)
+		glade_project_sync_resources_for_widget (project, NULL, gwidget, TRUE);
 }
 
 /**
@@ -569,7 +727,7 @@ glade_project_is_selected (GladeProject *project,
  * @project: a #GladeProject
  * @emit_signal: whether or not to emit a signal indication a selection change
  *
- * TODO: write me
+ * Clears @project's selection chain
  *
  * If @emit_signal is %TRUE, calls glade_project_selection_changed().
  */
@@ -592,10 +750,11 @@ glade_project_selection_clear (GladeProject *project, gboolean emit_signal)
 /**
  * glade_project_selection_remove:
  * @project: a #GladeProject
- * @widget:
- * @emit_signal: whether or not to emit a signal indication a selection change
+ * @object:  a #GObject in @project
+ * @emit_signal: whether or not to emit a signal 
+ *               indicating a selection change
  *
- * TODO: write me
+ * Removes @object from the selection chain of @project
  *
  * If @emit_signal is %TRUE, calls glade_project_selection_changed().
  */
@@ -620,10 +779,11 @@ glade_project_selection_remove (GladeProject *project,
 /**
  * glade_project_selection_add:
  * @project: a #GladeProject
- * @widget:
- * @emit_signal: whether or not to emit a signal indication a selection change
+ * @object:  a #GObject in @project
+ * @emit_signal: whether or not to emit a signal indicating 
+ *               a selection change
  *
- * TODO: write me
+ * Adds @object to the selection chain of @project
  *
  * If @emit_signal is %TRUE, calls glade_project_selection_changed().
  */
@@ -649,10 +809,11 @@ glade_project_selection_add (GladeProject *project,
 /**
  * glade_project_selection_set:
  * @project: a #GladeProject
- * @widget:
- * @emit_signal: whether or not to emit a signal indication a selection change
+ * @object:  a #GObject in @project
+ * @emit_signal: whether or not to emit a signal 
+ *               indicating a selection change
  *
- * TODO: write me
+ * Set the selection in @project to @object
  *
  * If @emit_signal is %TRUE, calls glade_project_selection_changed().
  */
@@ -690,6 +851,37 @@ glade_project_selection_get (GladeProject *project)
 	return project->selection;
 }
 
+static GList *
+glade_project_required_libs (GladeProject *project)
+{
+	GList       *required = NULL, *l, *ll;
+	GladeWidget *gwidget;
+	gboolean     listed;
+
+	for (l = project->objects; l; l = l->next)
+	{
+		gwidget = glade_widget_get_from_gobject (l->data);
+		g_assert (gwidget);
+
+		if (gwidget->widget_class->catalog)
+		{
+			listed = FALSE;
+			for (ll = required; ll; ll = ll->next)
+				if (!strcmp ((gchar *)ll->data, 
+					     gwidget->widget_class->catalog))
+				{
+					listed = TRUE;
+					break;
+				}
+
+			if (!listed)
+				required = g_list_prepend 
+					(required, gwidget->widget_class->catalog);
+		}
+	}
+	return required;
+}
+
 /**
  * glade_project_write:
  * @project: a #GladeProject
@@ -698,20 +890,28 @@ glade_project_selection_get (GladeProject *project)
  *          project and its contents
  */
 static GladeInterface *
-glade_project_write (const GladeProject *project)
+glade_project_write (GladeProject *project)
 {
-	GladeInterface *interface;
-	GList *list, *tops = NULL;
-	guint i;
+	GladeInterface  *interface;
+	GList           *required, *list, *tops = NULL;
+	gchar          **strv = NULL;
+	guint            i;
 
-	interface = g_new0 (GladeInterface, 1);
-	interface->names = g_hash_table_new (g_str_hash, g_str_equal);
-	interface->strings = g_hash_table_new_full (g_str_hash,
-						    g_str_equal,
-						    (GDestroyNotify)g_free,
-						    NULL);
+	interface = glade_interface_new ();
 
-        for (i = 0, list = project->objects; list; list = list->next)
+	if ((required = glade_project_required_libs (project)) != NULL)
+	{
+		strv = g_malloc0 (g_list_length (required) * sizeof (char *));
+		for (i = 0, list = required; list; i++, list = list->next)
+			strv[i] = g_strdup (list->data);
+
+		g_list_free (required);
+
+		interface->n_requires = g_list_length (required);
+		interface->requires   = strv;
+	}
+
+	for (i = 0, list = project->objects; list; list = list->next)
 	{
 		GladeWidget *widget;
 		GladeWidgetInfo *info;
@@ -742,25 +942,62 @@ glade_project_write (const GladeProject *project)
 	return interface;
 }
 
+static gboolean
+glade_project_set_changed_to_false_idle (gpointer data)
+{
+	((GladeProject *)data)->changed = FALSE;
+	return FALSE;
+}
+
+static gboolean
+loadable_interface (GladeInterface *interface, const gchar *path)
+{
+	GString *string = g_string_new (NULL);
+	gboolean loadable = TRUE;
+	guint i;
+
+	/* Check for required plugins here
+	 */
+	for (i = 0; i < interface->n_requires; i++)
+		if (!glade_catalog_is_loaded (interface->requires[i]))
+		{
+			g_string_append (string, interface->requires[i]);
+			loadable = FALSE;
+		}
+
+	if (loadable == FALSE)
+		glade_util_ui_message (glade_default_app_get_window(),
+				       GLADE_UI_ERROR,
+				       _("Failed to load %s.\n"
+					 "The following required catalogs are unavailable: %s"),
+				       path, string->str);
+	g_string_free (string, TRUE);
+	return loadable;
+}
+
 static GladeProject *
 glade_project_new_from_interface (GladeInterface *interface, const gchar *path)
 {
 	GladeProject *project;
-	GladeWidget *widget;
+	GladeWidget  *widget;
 	guint i;
 
 	g_return_val_if_fail (interface != NULL, NULL);
+	g_return_val_if_fail (path != NULL, NULL);
+
+	if (loadable_interface (interface, path) == FALSE)
+		return NULL;
 
 	project = glade_project_new (FALSE);
-	project->path = g_strdup (path);
-	g_free (project->name);
+
+	project->path = glade_util_canonical_path (path);
+	
+	if (project->name) g_free (project->name);
+
 	project->name = g_path_get_basename (path);
 	project->selection = NULL;
 	project->objects = NULL;
 
-	if (interface->n_requires)
-		g_warning ("We currently do not support projects requiring additional libs");
-
 	for (i = 0; i < interface->n_toplevels; ++i)
 	{
 		widget = glade_widget_read ((gpointer)project, interface->toplevels[i]);
@@ -769,14 +1006,48 @@ glade_project_new_from_interface (GladeInterface *interface, const gchar *path)
 			g_warning ("Failed to read a <widget> tag");
 			continue;
 		}
-		glade_project_add_object (project, widget->object);
+		glade_project_add_object (project, NULL, widget->object);
 	}
 
-	project->changed = FALSE;
+	/* Set project status after every idle functions */
+	g_idle_add_full (G_PRIORITY_LOW,
+			 glade_project_set_changed_to_false_idle,
+			 project, 
+			 NULL);
 
 	return project;	
 }
 
+static void 
+glade_project_fix_object_props (GladeProject *project)
+{
+	GList         *l, *ll;
+	GladeWidget   *gwidget, *value_widget;
+	GladeProperty *property;
+	gchar         *txt;
+
+	for (l = project->objects; l; l = l->next)
+	{
+		gwidget = glade_widget_get_from_gobject (l->data);
+
+		for (ll = gwidget->properties; ll; ll = ll->next)
+		{
+			property = GLADE_PROPERTY (ll->data);
+
+			if (glade_property_class_is_object (property->class) &&
+			    (txt = g_object_get_data (G_OBJECT (property), 
+						      "glade-loaded-object")) != NULL)
+			{
+				if ((value_widget = 
+				     glade_project_get_widget_by_name (project, txt)) != NULL)
+					glade_property_set (property, value_widget->object);
+
+				g_object_set_data (G_OBJECT (property), "glade-loaded-object", NULL);
+			}
+		}
+	}
+}
+
 /**
  * glade_project_open:
  * @path:
@@ -789,22 +1060,58 @@ glade_project_new_from_interface (GladeInterface *interface, const gchar *path)
 GladeProject *
 glade_project_open (const gchar *path)
 {
-	GladeProject *project;
+	GladeProject *project = NULL;
 	GladeInterface *interface;
 	
 	g_return_val_if_fail (path != NULL, NULL);
 
-	interface = glade_parser_parse_file (path, NULL);
-	if (!interface)
-		return NULL;
-
-	project = glade_project_new_from_interface (interface, path);
-	
-	glade_interface_destroy (interface);
-
+	if ((interface = 
+	     glade_parser_parse_file (path, NULL)) != NULL)
+	{
+		if ((project = 
+		     glade_project_new_from_interface (interface, 
+						       path)) != NULL)
+		{
+			/* Now we have to loop over all the object properties
+			 * and fix'em all ('cause they probably weren't found)
+			 */
+			glade_project_fix_object_props (project);
+		}
+		
+		glade_interface_destroy (interface);
+	}
 	return project;
 }
 
+static void
+glade_project_move_resources (GladeProject *project,
+			      const gchar  *old_dir,
+			      const gchar  *new_dir)
+{
+	GList *list, *resources;
+	gchar *old_name, *new_name;
+
+	if (old_dir == NULL || /* <-- Cant help you :( */
+	    new_dir == NULL)   /* <-- Unlikely         */
+		return;
+
+	if ((resources = /* Nothing to do here */
+	     glade_project_list_resources (project)) == NULL)
+		return;
+	
+	for (list = resources; list; list = list->next)
+	{
+		old_name = g_build_filename 
+			(old_dir, (gchar *)list->data, NULL);
+		new_name = g_build_filename 
+			(new_dir, (gchar *)list->data, NULL);
+		glade_util_copy_file (old_name, new_name);
+		g_free (old_name);
+		g_free (new_name);
+	}
+	g_list_free (resources);
+}
+
 /**
  * glade_project_save:
  * @project: a #GladeProject
@@ -820,6 +1127,7 @@ glade_project_save (GladeProject *project, const gchar *path, GError **error)
 {
 	GladeInterface *interface;
 	gboolean        ret;
+	gchar          *canonical_path;
 
 	g_return_val_if_fail (GLADE_IS_PROJECT (project), FALSE);
 
@@ -833,19 +1141,43 @@ glade_project_save (GladeProject *project, const gchar *path, GError **error)
 	ret = glade_interface_dump_full (interface, path, error);
 	glade_interface_destroy (interface);
 
-	if (path != project->path)
-        {
-		g_free (project->path);
-		project->path = g_strdup_printf ("%s", path);
-	}
-	g_free (project->name);
-	project->name = g_path_get_basename (project->path);
+	canonical_path = glade_util_canonical_path (path);
 
+	if (strcmp (canonical_path, project->path))
+        {
+		gchar *old_dir, *new_dir;
+
+		if (project->path)
+		{
+			old_dir = g_path_get_dirname (project->path);
+			new_dir = g_path_get_dirname (canonical_path);
+
+			glade_project_move_resources (project, 
+						      old_dir, 
+						      new_dir);
+			g_free (old_dir);
+			g_free (new_dir);
+		}
+		project->path = (g_free (project->path),
+				 g_strdup (canonical_path));
+
+		project->name = (g_free (project->name),
+				 g_path_get_basename (project->path));
+	}
+
+	g_free (canonical_path);
 	project->changed = FALSE;
 
 	return ret;
 }
 
+void
+glade_project_reset_path (GladeProject *project)
+{
+	g_return_if_fail (GLADE_IS_PROJECT (project));
+	project->path = (g_free (project->path), NULL);
+}
+
 void
 glade_project_changed (GladeProject *project)
 {
@@ -868,32 +1200,6 @@ glade_project_get_tooltips (GladeProject *project)
 }
 
 /**
- * glade_project_get_menuitem
- */
-GtkWidget *
-glade_project_get_menuitem (GladeProject *project)
-{
-	GtkUIManager *ui;
-	gchar        *path;
-
-	g_return_val_if_fail (GLADE_IS_PROJECT (project), NULL);
-
-	ui   = g_object_get_data (G_OBJECT (project->action), "ui");
-	path = g_object_get_data (G_OBJECT (project->action), "menuitem_path");
-	return gtk_ui_manager_get_widget (ui, path);
-}
-
-/**
- * glade_project_get_menuitem_merge_id
- */
-guint
-glade_project_get_menuitem_merge_id (GladeProject *project)
-{
-	g_return_val_if_fail (GLADE_IS_PROJECT (project), 0);
-	return GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (project->action), "merge_id"));
-}
-
-/*
  * glade_project_set_accel_group:
  *
  * Set @accel_group to every top level widget in @project.
@@ -909,8 +1215,159 @@ glade_project_set_accel_group (GladeProject *project, GtkAccelGroup *accel_group
 	while (objects)
 	{
 		if(GTK_IS_WINDOW (objects->data))
+		{
+			if (project->accel_group)
+				gtk_window_remove_accel_group (GTK_WINDOW (objects->data), project->accel_group);
+			
 			gtk_window_add_accel_group (GTK_WINDOW (objects->data), accel_group);
+		}
 
 		objects = objects->next;
 	}
+	
+	project->accel_group = accel_group;
+}
+
+/**
+ * glade_project_resource_fullpath:
+ * @project: The #GladeProject.
+ * @resource: The resource basename
+ *
+ * Returns: A newly allocated string holding the 
+ *          full path the the project resource.
+ */
+gchar *
+glade_project_resource_fullpath (GladeProject *project,
+				 const gchar  *resource)
+{
+	gchar *fullpath, *project_dir;
+
+	g_return_val_if_fail (GLADE_IS_PROJECT (project), NULL);
+
+	project_dir = g_path_get_dirname (project->path);
+	fullpath    = g_build_filename (project_dir, resource, NULL);
+	g_free (project_dir);
+
+	return fullpath;
+}
+
+
+/**
+ * glade_project_set_resource:
+ * @project: A #GladeProject
+ * @property: The #GladeProperty this resource is required by
+ * @resource: The resource file basename to be found in the same
+ *            directory as the glade file.
+ *
+ * Adds/Modifies/Removes a resource from a project; any project resources
+ * will be copied when using "Save As...", when moving widgets across projects
+ * and will be copied into the project's directory when selected.
+ */
+void
+glade_project_set_resource (GladeProject  *project, 
+			    GladeProperty *property,
+			    const gchar   *resource)
+{
+	gchar *last_resource, *last_resource_dup = NULL, *base_resource = NULL;
+	gchar *fullpath, *dirname;
+
+	g_return_if_fail (GLADE_IS_PROJECT (project));
+	g_return_if_fail (GLADE_IS_PROPERTY (property));
+
+	last_resource = g_hash_table_lookup (project->resources, property);
+	if (resource) base_resource = g_path_get_basename (resource);
+
+	if (last_resource)
+		last_resource_dup = g_strdup (last_resource);
+	
+	if (last_resource_dup && 
+	    (base_resource == NULL || strcmp (last_resource_dup, base_resource)))
+	{
+		g_hash_table_remove (project->resources, property);
+
+		/* Emit remove signal
+		 */
+		g_signal_emit (G_OBJECT (project),
+			       glade_project_signals [RESOURCE_REMOVED],
+			       0, last_resource_dup);
+	}
+	
+	if (project->path)
+	{
+		dirname = g_path_get_dirname (project->path);
+		fullpath = g_build_filename (dirname, base_resource, NULL);
+	
+		if (resource && project->path && 
+		    g_file_test (resource, G_FILE_TEST_IS_REGULAR) &&
+		    strcmp (fullpath, resource))
+		{
+			glade_util_copy_file (resource, fullpath);
+		}
+		g_free (fullpath);
+		g_free (dirname);
+	}
+
+	g_hash_table_insert (project->resources, property, base_resource);
+
+	/* Emit update signal
+	 */
+	if (base_resource)
+		g_signal_emit (G_OBJECT (project),
+			       glade_project_signals [RESOURCE_UPDATED],
+			       0, base_resource);
+}
+
+static void
+list_resources_accum (GladeProperty  *key,
+		      gchar          *value,
+		      GList         **list)
+{
+	*list = g_list_prepend (*list, value);
+}
+
+
+
+/**
+ * glade_project_list_resources:
+ * @project: A #GladeProject
+ *
+ * Returns: A newly allocated #GList of file basenames
+ *          of resources in this project, note that the
+ *          strings are not allocated and are unsafe to
+ *          use once the projects state changes.
+ *          The returned list should be freed with g_list_free.
+ */
+GList *
+glade_project_list_resources (GladeProject  *project)
+{
+	GList *list = NULL;
+	g_return_val_if_fail (GLADE_IS_PROJECT (project), NULL);
+
+	g_hash_table_foreach (project->resources, 
+			      (GHFunc)list_resources_accum, &list);
+	return list;
+}
+
+
+
+/**
+ * glade_project_display_name:
+ * @project: A #GladeProject
+ *
+ * Returns: A newly allocated string to uniquely 
+ *          describe this open project.
+ *       
+ */
+gchar *
+glade_project_display_name (GladeProject *project)
+{
+	gchar *final_name = NULL;
+	if (project->instance > 0)
+		final_name = g_strdup_printf ("%s <%d>", 
+					      project->name, 
+					      project->instance);
+	else
+		final_name = g_strdup (project->name);
+
+	return final_name;
 }
