@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * Authors:
  *   Chema Celorio <chema@celorio.com>
@@ -108,6 +108,14 @@ glade_property_default_impl (GladeProperty *property)
 				    property->class->def);
 }
 
+gboolean
+glade_property_equals_value_impl (GladeProperty *property,
+				  const GValue  *value)
+{
+	return !g_param_values_cmp (property->class->pspec,
+				    property->value, value);
+}
+
 /**
  * Generic set function for properties that do not have a
  * custom set_property method. This includes packing properties.
@@ -155,6 +163,27 @@ glade_property_set_value_impl (GladeProperty *property, const GValue *value)
 	changed = g_param_values_cmp (property->class->pspec, 
 				      property->value, value) != 0;
 
+
+	/* Add/Remove references from widget ref stacks here
+	 */
+	if (changed && glade_property_class_is_object (property->class))
+	{
+		GladeWidget *gold, *gnew;
+		GObject     *old_object, *new_object;
+
+		if ((old_object = g_value_get_object (property->value)) != NULL)
+		{
+			gold = glade_widget_get_from_gobject (old_object);
+			glade_widget_remove_prop_ref (gold, property);
+		}
+
+		if ((new_object = g_value_get_object (value)) != NULL)
+		{
+			gnew = glade_widget_get_from_gobject (new_object);
+			glade_widget_add_prop_ref (gnew, property);
+		}
+	}
+
 	/* Assign property first so that; if the object need be
 	 * rebuilt, it will reflect the new value
 	 */
@@ -195,10 +224,10 @@ glade_property_sync_impl (GladeProperty *property)
 {
 	if (property->enabled == FALSE || 
 	    property->class->ignore    ||
-	    property->loading)
+	    property->syncing)
 		return;
 
-	property->loading = TRUE;
+	property->syncing = TRUE;
 
 	if (property->class->set_function)
 		/* if there is a custom set_property, use it */
@@ -240,7 +269,7 @@ glade_property_sync_impl (GladeProperty *property)
 
 		if (inproject)
 		{
-			glade_project_add_object (property->widget->project,
+			glade_project_add_object (property->widget->project, NULL,
 						  glade_widget_get_object (property->widget));
 			if (reselect)
 				glade_project_selection_add
@@ -251,7 +280,7 @@ glade_property_sync_impl (GladeProperty *property)
 	else
 		glade_property_set_property (property, property->value);
 
-	property->loading = FALSE;
+	property->syncing = FALSE;
 }
 
 static gboolean
@@ -276,17 +305,6 @@ glade_property_write_impl (GladeProperty  *property,
 
 	/* put the name="..." part on the <property ...> tag */
 	info.name = alloc_propname(interface, tmp);
- 	/*if (state->translate_prop && state->content->str[0] != '\0') {
-	    if (state->context_prop) 
-		info.value = alloc_string(state->interface,
-					  g_strip_context(state->content->str,
-							   dgettext(state->domain, state->content->str)));
-	    else
-		info.value = alloc_string(state->interface,
-					  dgettext(state->domain, state->content->str));
- 	} else {
- 	    info.value = alloc_string(state->interface, state->content->str);
-  	}*/
 	g_free (tmp);
 
 	/* convert the value of this property to a string, and put it between
@@ -304,6 +322,9 @@ glade_property_write_impl (GladeProperty  *property,
 	}
 	else
 	{
+		/* Skip properties that are default
+		 * (by original pspec default) 
+		 */
 		default_str =
 			glade_property_class_make_string_from_gvalue
 			(property->class, property->class->orig_def);
@@ -316,10 +337,14 @@ glade_property_write_impl (GladeProperty  *property,
 		g_free (tmp);
 		return FALSE;
 	}
-	
+
 	if (property->class->translatable)
 	{
-		/* FIXME: implement writing the i18n metadata. */
+		info.translatable = property->i18n_translatable;
+		info.has_context  = property->i18n_has_context;
+		if (property->i18n_comment)
+			info.comment = alloc_string
+				(interface, property->i18n_comment);
 	}
 	
 	info.value = alloc_string(interface, tmp);
@@ -427,7 +452,7 @@ glade_property_init (GladeProperty *property)
 {
 	property->enabled   = TRUE;
 	property->sensitive = TRUE;
-	property->i18n_translatable = FALSE;
+	property->i18n_translatable = TRUE;
 	property->i18n_has_context = FALSE;
 	property->i18n_comment = NULL;
 }
@@ -450,6 +475,7 @@ glade_property_klass_init (GladePropertyKlass *prop_class)
 	prop_class->dup                   = glade_property_dup_impl;
 	prop_class->reset                 = glade_property_reset_impl;
 	prop_class->def                   = glade_property_default_impl;
+	prop_class->equals_value          = glade_property_equals_value_impl;
 	prop_class->set_value             = glade_property_set_value_impl;
 	prop_class->get_value             = glade_property_get_value_impl;
 	prop_class->get_default           = glade_property_get_default_impl;
@@ -472,7 +498,7 @@ glade_property_klass_init (GladePropertyKlass *prop_class)
 		 g_param_spec_boolean 
 		 ("sensitive", _("Sensitive"), 
 		  _("This gives backends control to set property sensitivity"),
-		  TRUE, G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
+		  TRUE, G_PARAM_READWRITE));
 
 	g_object_class_install_property 
 		(object_class, PROP_I18N_COMMENT,
@@ -552,12 +578,13 @@ glade_property_get_type (void)
  *******************************************************************************/
 /**
  * glade_property_new:
- * @class:
- * @widget:
+ * @class: A #GladeWidgetClass
+ * @widget: A #GladeWidget
+ * @value: A #GValue
  *
- * TODO: write me
  *
- * Returns:
+ * Returns: A newly created #GladeProperty based on the
+ *          given criteria
  */
 GladeProperty *
 glade_property_new (GladePropertyClass *class, 
@@ -596,12 +623,10 @@ glade_property_new (GladePropertyClass *class,
 
 /**
  * glade_property_dup:
- * @template:
- * @widget:
+ * @template: A #GladeProperty
+ * @widget: A #GladeWidget
  *
- * TODO: write me
- *
- * Returns:
+ * Returns: A newly duplicated property based on the new widget
  */
 GladeProperty *
 glade_property_dup (GladeProperty *template, GladeWidget *widget)
@@ -612,9 +637,9 @@ glade_property_dup (GladeProperty *template, GladeWidget *widget)
 
 /**
  * glade_property_reset:
- * @property:
+ * @property: A #GladeProperty
  *
- * TODO: write me
+ * Resets this property to its default value
  */
 void
 glade_property_reset (GladeProperty *property)
@@ -624,10 +649,10 @@ glade_property_reset (GladeProperty *property)
 }
 
 /**
- * glade_property_reset:
- * @property:
+ * glade_property_default:
+ * @property: A #GladeProperty
  *
- * TODO: write me
+ * Returns: Whether this property is at its default value
  */
 gboolean
 glade_property_default (GladeProperty *property)
@@ -636,12 +661,73 @@ glade_property_default (GladeProperty *property)
 	return GLADE_PROPERTY_GET_KLASS (property)->def (property);
 }
 
+/**
+ * glade_property_equals_value:
+ * @property: a #GladeProperty
+ * @value: a #GValue
+ *
+ * Returns: Whether this property is equal to the value provided
+ */
+gboolean
+glade_property_equals_value (GladeProperty      *property, 
+			     GValue             *value)
+{
+	g_return_val_if_fail (GLADE_IS_PROPERTY (property), FALSE);
+	return GLADE_PROPERTY_GET_KLASS (property)->equals_value (property, value);
+}
+
+/**
+ * glade_property_equals_va_list:
+ * @property: a #GladeProperty
+ * @vl: a va_list
+ *
+ * Returns: Whether this property is equal to the value provided
+ */
+gboolean
+glade_property_equals_va_list (GladeProperty *property, va_list vl)
+{
+	GValue   *value;
+	gboolean  ret;
+
+	g_return_val_if_fail (GLADE_IS_PROPERTY (property), FALSE);
+
+	value = glade_property_class_make_gvalue_from_vl (property->class, vl);
+
+	ret = GLADE_PROPERTY_GET_KLASS (property)->equals_value (property, value);
+	
+	g_value_unset (value);
+	g_free (value);
+	return ret;
+}
+
+/**
+ * glade_property_equals:
+ * @property: a #GladeProperty
+ * @...: a provided property value
+ *
+ * Returns: Whether this property is equal to the value provided
+ */
+gboolean
+glade_property_equals (GladeProperty *property, ...)
+{
+	va_list  vl;
+	gboolean ret;
+
+	g_return_val_if_fail (GLADE_IS_PROPERTY (property), FALSE);
+
+	va_start (vl, property);
+	ret = glade_property_equals_va_list (property, vl);
+	va_end (vl);
+
+	return ret;
+}
+
 /**
  * glade_property_set_value:
  * @property: a #GladeProperty
  * @value: a #GValue
  *
- * TODO: write me
+ * Sets the property's value
  */
 void
 glade_property_set_value (GladeProperty *property, const GValue *value)
@@ -656,7 +742,7 @@ glade_property_set_value (GladeProperty *property, const GValue *value)
  * @property: a #GladeProperty
  * @va_list: a va_list with value to set
  *
- * TODO: write me
+ * Sets the property's value
  */
 void
 glade_property_set_va_list (GladeProperty *property, va_list vl)
@@ -665,42 +751,7 @@ glade_property_set_va_list (GladeProperty *property, va_list vl)
 
 	g_return_if_fail (GLADE_IS_PROPERTY (property));
 
-	value = g_new0 (GValue, 1);
-	g_value_init (value, property->class->pspec->value_type);
-	
-	if (G_IS_PARAM_SPEC_ENUM(property->class->pspec))
-		g_value_set_enum (value, va_arg (vl, gint));
-	else if (G_IS_PARAM_SPEC_FLAGS(property->class->pspec))
-		g_value_set_flags (value, va_arg (vl, gint));
-	else if (G_IS_PARAM_SPEC_INT(property->class->pspec))
-		g_value_set_int (value, va_arg (vl, gint));
-	else if (G_IS_PARAM_SPEC_UINT(property->class->pspec))
-		g_value_set_uint (value, va_arg (vl, guint));
-	else if (G_IS_PARAM_SPEC_LONG(property->class->pspec))
-		g_value_set_long (value, va_arg (vl, glong));
-	else if (G_IS_PARAM_SPEC_ULONG(property->class->pspec))
-		g_value_set_ulong (value, va_arg (vl, gulong));
-	else if (G_IS_PARAM_SPEC_INT64(property->class->pspec))
-		g_value_set_int64 (value, va_arg (vl, gint64));
-	else if (G_IS_PARAM_SPEC_UINT64(property->class->pspec))
-		g_value_set_uint64 (value, va_arg (vl, guint64));
-	else if (G_IS_PARAM_SPEC_FLOAT(property->class->pspec))
-		g_value_set_float (value, (gfloat)va_arg (vl, gdouble));
-	else if (G_IS_PARAM_SPEC_DOUBLE(property->class->pspec))
-		g_value_set_double (value, va_arg (vl, gdouble));
-	else if (G_IS_PARAM_SPEC_STRING(property->class->pspec))
-		g_value_set_string (value, va_arg (vl, gchar *));
-	else if (G_IS_PARAM_SPEC_CHAR(property->class->pspec))
-		g_value_set_char (value, (gchar)va_arg (vl, gint));
-	else if (G_IS_PARAM_SPEC_UCHAR(property->class->pspec))
-		g_value_set_uchar (value, (guchar)va_arg (vl, guint));
-	else if (G_IS_PARAM_SPEC_UNICHAR(property->class->pspec))
-		g_value_set_uint (value, va_arg (vl, gunichar));
-	else if (G_IS_PARAM_SPEC_BOOLEAN(property->class->pspec))
-		g_value_set_boolean (value, va_arg (vl, gboolean));
-	else
-		g_critical ("Unsupported pspec type %s",
-			    g_type_name(property->class->pspec->value_type));
+	value = glade_property_class_make_gvalue_from_vl (property->class, vl);
 
 	GLADE_PROPERTY_GET_KLASS (property)->set_value (property, value);
 	
@@ -713,7 +764,7 @@ glade_property_set_va_list (GladeProperty *property, va_list vl)
  * @property: a #GladeProperty
  * @...: the value to set
  *
- * TODO: write me
+ * Sets the property's value (in a convenient way)
  */
 void
 glade_property_set (GladeProperty *property, ...)
@@ -732,7 +783,7 @@ glade_property_set (GladeProperty *property, ...)
  * @property: a #GladeProperty
  * @value: a #GValue
  *
- * TODO: write me
+ * Retrieve the property value
  */
 void
 glade_property_get_value (GladeProperty *property, GValue *value)
@@ -747,7 +798,7 @@ glade_property_get_value (GladeProperty *property, GValue *value)
  * @property: a #GladeProperty
  * @value: a #GValue
  *
- * TODO: write me
+ * Retrieve the default property value
  */
 void
 glade_property_get_default (GladeProperty *property, GValue *value)
@@ -760,67 +811,23 @@ glade_property_get_default (GladeProperty *property, GValue *value)
 /**
  * glade_property_get_va_list:
  * @property: a #GladeProperty
- * @value: a #GValue
+ * @vl: a va_list
  *
- * TODO: write me
+ * Retrieve the property value
  */
 void
 glade_property_get_va_list (GladeProperty *property, va_list vl)
 {
-	GValue  *value;
-
 	g_return_if_fail (GLADE_IS_PROPERTY (property));
-
-	value = g_new0 (GValue, 1);
-	GLADE_PROPERTY_GET_KLASS (property)->get_value (property, value);
-
-	/* The argument is a pointer of the specified type, cast the pointer and assign
-	 * the value using the proper g_value_get_ variation.
-	 */
-	if (G_IS_PARAM_SPEC_ENUM(property->class->pspec))
-		*(gint *)(va_arg (vl, gint *)) = g_value_get_enum (property->value);
-	else if (G_IS_PARAM_SPEC_FLAGS(property->class->pspec))
-		*(gint *)(va_arg (vl, gint *)) = g_value_get_flags (property->value);
-	else if (G_IS_PARAM_SPEC_INT(property->class->pspec))
-		*(gint *)(va_arg (vl, gint *)) = g_value_get_int (property->value);
-	else if (G_IS_PARAM_SPEC_UINT(property->class->pspec))
-		*(guint *)(va_arg (vl, guint *)) = g_value_get_uint (property->value);
-	else if (G_IS_PARAM_SPEC_LONG(property->class->pspec))
-		*(glong *)(va_arg (vl, glong *)) = g_value_get_long (property->value);
-	else if (G_IS_PARAM_SPEC_ULONG(property->class->pspec))
-		*(gulong *)(va_arg (vl, gulong *)) = g_value_get_ulong (property->value);
-	else if (G_IS_PARAM_SPEC_INT64(property->class->pspec))
-		*(gint64 *)(va_arg (vl, gint64 *)) = g_value_get_int64 (property->value);
-	else if (G_IS_PARAM_SPEC_UINT64(property->class->pspec))
-		*(guint64 *)(va_arg (vl, guint64 *)) = g_value_get_uint64 (property->value);
-	else if (G_IS_PARAM_SPEC_FLOAT(property->class->pspec))
-		*(gfloat *)(va_arg (vl, gdouble *)) = g_value_get_float (property->value);
-	else if (G_IS_PARAM_SPEC_DOUBLE(property->class->pspec))
-		*(gdouble *)(va_arg (vl, gdouble *)) = g_value_get_double (property->value);
-	else if (G_IS_PARAM_SPEC_STRING(property->class->pspec))
-		*(gchar **)(va_arg (vl, gchar *)) = (gchar *)g_value_get_string (property->value);
-	else if (G_IS_PARAM_SPEC_CHAR(property->class->pspec))
-		*(gchar *)(va_arg (vl, gint *)) = g_value_get_char (property->value);
-	else if (G_IS_PARAM_SPEC_UCHAR(property->class->pspec))
-		*(guchar *)(va_arg (vl, guint *)) = g_value_get_uchar (property->value);
-	else if (G_IS_PARAM_SPEC_UNICHAR(property->class->pspec))
-		*(guint *)(va_arg (vl, gunichar *)) = g_value_get_uint (property->value);
-	else if (G_IS_PARAM_SPEC_BOOLEAN(property->class->pspec))
-		*(gboolean *)(va_arg (vl, gboolean *)) = g_value_get_boolean (property->value);
-	else
-		g_critical ("Unsupported pspec type %s",
-			    g_type_name(property->class->pspec->value_type));
-
-	g_value_unset (value);
-	g_free (value);
+	glade_property_class_set_vl_from_gvalue (property->class, property->value, vl);
 }
 
 /**
  * glade_property_get:
  * @property: a #GladeProperty
- * @value: a #GValue
+ * @...: An address to store the value
  *
- * TODO: write me
+ * Retrieve the property value
  */
 void
 glade_property_get (GladeProperty *property, ...)
@@ -838,7 +845,7 @@ glade_property_get (GladeProperty *property, ...)
  * glade_property_sync:
  * @property: a #GladeProperty
  *
- * TODO: write me
+ * Synchronize the object with this property
  */
 void
 glade_property_sync (GladeProperty *property)
@@ -853,9 +860,7 @@ glade_property_sync (GladeProperty *property)
  * @interface: a #GladeInterface
  * @props: a GArray of #GladePropInfo
  *
- * TODO: write me
- *
- * Returns:
+ * Write this property to the GladeInterface metadata
  */
 gboolean
 glade_property_write (GladeProperty *property, GladeInterface *interface, GArray *props)
