@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * Authors:
  *   Chema Celorio <chema@celorio.com>
@@ -41,7 +41,6 @@
 #include "glade-xml-utils.h"
 #include "glade-property.h"
 #include "glade-property-class.h"
-#include "glade-catalog.h"
 #include "glade-signal.h"
 #include "glade-parameter.h"
 #include "glade-debug.h"
@@ -95,6 +94,7 @@ glade_widget_class_free (GladeWidgetClass *widget_class)
 
 	g_free (widget_class->generic_name);
 	g_free (widget_class->name);
+	g_free (widget_class->catalog);
 
 	g_list_foreach (widget_class->properties, (GFunc) glade_property_class_free, NULL);
 	g_list_free (widget_class->properties);
@@ -116,32 +116,13 @@ glade_widget_class_free (GladeWidgetClass *widget_class)
 }
 
 static GList *
-glade_widget_class_list_properties (GladeWidgetClass *class)
+gwc_props_from_pspecs (GParamSpec **specs, gint n_specs)
 {
 	GladePropertyClass *property_class;
-	GObjectClass *object_class;
-	GParamSpec **specs = NULL;
-	GParamSpec *spec;
-	GType last;
-	guint n_specs = 0;
-	gint i;
-	GList *list = NULL;
+	GParamSpec         *spec;
+	gint                i;
+	GList              *list = NULL;
 
-	g_return_val_if_fail (GLADE_IS_WIDGET_CLASS (class), NULL);
-
-	/* Let it leek */
-	g_type_class_ref (class->type);
-
-	object_class = g_type_class_peek (class->type);
-	if (!object_class)
-	{
-		g_warning ("Class peek failed\n");
-		return NULL;
-	}
-
-	specs = g_object_class_list_properties (object_class, &n_specs);
-
-	last = 0;
 	for (i = 0; i < n_specs; i++)
 	{
 		spec = specs[i];
@@ -153,8 +134,6 @@ glade_widget_class_list_properties (GladeWidgetClass *class)
 			if (!property_class)
 				continue;
 
-			property_class->optional = FALSE;
-
 			/* Flag the construct only properties */
 			if (spec->flags & G_PARAM_CONSTRUCT_ONLY)
 				property_class->construct_only = TRUE;
@@ -162,9 +141,29 @@ glade_widget_class_list_properties (GladeWidgetClass *class)
 			list = g_list_prepend (list, property_class);
 		}
 	}
+	return g_list_reverse (list);
+}
 
-	list = g_list_reverse (list);
+static GList *
+glade_widget_class_list_properties (GladeWidgetClass *class)
+{
+	GObjectClass  *object_class;
+	GParamSpec   **specs = NULL;
+	guint          n_specs = 0;
+	GList         *list;
 
+	g_return_val_if_fail (GLADE_IS_WIDGET_CLASS (class), NULL);
+
+	/* Let it leek */
+	if ((object_class = g_type_class_ref (class->type)) == NULL)
+	{
+		g_warning ("Failed to get class for type %s\n", 
+			   g_type_name (class->type));
+		return NULL;
+	}
+
+	specs = g_object_class_list_properties (object_class, &n_specs);
+	list = gwc_props_from_pspecs (specs, n_specs);
 	g_free (specs);
 
 	return list;
@@ -173,56 +172,36 @@ glade_widget_class_list_properties (GladeWidgetClass *class)
 static GList * 
 glade_widget_class_list_child_properties (GladeWidgetClass *class) 
 {
-	GladePropertyClass *property_class;
-	GObjectClass *object_class;
-	GParamSpec **specs = NULL;
-	GParamSpec *spec;
-	guint n_specs = 0;
-	gint i;
-	GList *list = NULL;
+	GladePropertyClass  *property_class;
+	GObjectClass        *object_class;
+	GParamSpec         **specs = NULL;
+	guint                n_specs = 0;
+	GList               *list = NULL, *l;
 
 	g_return_val_if_fail (GLADE_IS_WIDGET_CLASS (class), NULL);
 
-	/* only containers have child propeties */
+	/* only GtkContainer child propeties can be introspected */
 	if (!g_type_is_a (class->type, GTK_TYPE_CONTAINER))
 		return NULL;
 
-	object_class = g_type_class_peek (class->type);
-	if (!object_class)
+	/* Let it leek */
+	if ((object_class = g_type_class_ref (class->type)) == NULL)
 	{
-		g_warning ("Class peek failed\n");
+		g_warning ("Failed to get class for type %s\n", 
+			   g_type_name (class->type));
 		return NULL;
 	}
 
 	specs = gtk_container_class_list_child_properties (object_class, &n_specs);
-
-	for (i = 0; i < n_specs; i++)
-	{
-		spec = specs[i];
-
-		/* We only use the writable properties */
-		if (spec->flags & G_PARAM_WRITABLE)
-		{
-			property_class = glade_property_class_new_from_spec (spec);
-			if (!property_class)
-				continue;
-
-			property_class->optional = FALSE;
-			property_class->packing = TRUE;
-
-			/* Flag the construct only properties */
-			if (spec->flags & G_PARAM_CONSTRUCT_ONLY)
-				property_class->construct_only = TRUE;
-
-			
-			list = g_list_prepend (list, property_class);
-		}
-	}
-
-	list = g_list_reverse (list);
-
+	list  = gwc_props_from_pspecs (specs, n_specs);
 	g_free (specs);
 
+	/* Mark packing props */
+	for (l = list; l; l = l->next)
+	{
+		property_class = l->data;
+		property_class->packing = TRUE;
+	}
 	return list;
 }
 
@@ -335,9 +314,10 @@ glade_widget_class_create_icon (GladeWidgetClass *class)
 }
 
 static void
-glade_widget_class_update_properties_from_node (GladeXmlNode *node,
-						GladeWidgetClass *widget_class,
-						GList **properties)
+glade_widget_class_update_properties_from_node (GladeXmlNode      *node,
+						GladeWidgetClass  *widget_class,
+						GList            **properties,
+						const gchar       *domain)
 {
 	GladeXmlNode *child;
 
@@ -381,7 +361,7 @@ glade_widget_class_update_properties_from_node (GladeXmlNode *node,
 		updated = glade_property_class_update_from_node
 			(child, widget_class->module, 
 			 widget_class->type,
-			 &property_class);
+			 &property_class, domain);
 		if (!updated)
 		{
 			g_warning ("failed to update %s property of %s from xml",
@@ -483,7 +463,8 @@ glade_widget_class_load_function (GladeXmlNode     *node,
 
 static void
 glade_widget_class_update_children_from_node (GladeXmlNode     *node,
-					      GladeWidgetClass *widget_class)
+					      GladeWidgetClass *widget_class,
+					      const gchar      *domain)
 {
 	GladeSupportedChild *child;
 	GladeXmlNode        *child_node, *properties;
@@ -567,7 +548,7 @@ glade_widget_class_update_children_from_node (GladeXmlNode     *node,
 		     glade_xml_search_child (child_node, GLADE_TAG_PROPERTIES)) != NULL)
 		{
 			glade_widget_class_update_properties_from_node
-				(properties, widget_class, &child->properties);
+				(properties, widget_class, &child->properties, domain);
 
 			for (list = child->properties; list != NULL; list = list->next)
 				((GladePropertyClass *)list->data)->packing = TRUE;
@@ -577,7 +558,8 @@ glade_widget_class_update_children_from_node (GladeXmlNode     *node,
 
 static gboolean
 glade_widget_class_extend_with_node (GladeWidgetClass *widget_class, 
-				     GladeXmlNode     *node)
+				     GladeXmlNode     *node,
+				     const gchar      *domain)
 {
 	GladeXmlNode *child;
 
@@ -595,9 +577,13 @@ glade_widget_class_extend_with_node (GladeWidgetClass *widget_class,
 						  (void **)&widget_class->get_internal_child);
 
 		glade_widget_class_load_function (node, widget_class->module,
-						  GLADE_TAG_CHILD_PROPERTY_APPLIES_FUNCTION,
+						  GLADE_TAG_GET_INTERNAL_CHILDREN_FUNCTION,
+						  (void **)&widget_class->get_internal_children);
+
+		glade_widget_class_load_function (node, widget_class->module,
+						  GLADE_TAG_LAUNCH_EDITOR_FUNCTION,
 						  (void **)
-						  &widget_class->child_property_applies);
+						  &widget_class->launch_editor);
 	}
 
 	/* if we found a <properties> tag on the xml file, we add the properties
@@ -606,16 +592,15 @@ glade_widget_class_extend_with_node (GladeWidgetClass *widget_class,
 	child = glade_xml_search_child (node, GLADE_TAG_PROPERTIES);
 	if (child)
 	{
-		glade_widget_class_update_properties_from_node (child,
-								widget_class,
-								&widget_class->properties);
+		glade_widget_class_update_properties_from_node
+			(child, widget_class, &widget_class->properties, domain);
 	}
 
 	child = glade_xml_search_child (node, GLADE_TAG_CHILDREN);
 	if (child)
 	{
-		glade_widget_class_update_children_from_node (child,
-							      widget_class);
+		glade_widget_class_update_children_from_node
+			(child, widget_class, domain);
 	}
 
 	child = glade_xml_search_child (node, GLADE_TAG_PACKING_DEFAULTS);
@@ -752,11 +737,10 @@ glade_widget_class_merge_properties (GType   parent_type,
 		}
 	}
 
-
 	/* Remove any properties found in widget_properties not found in parent_properties
 	 * if parent_properties should have it through introspection
 	 */
-	object_class = g_type_class_peek (parent_type);
+	object_class = g_type_class_ref (parent_type); /* Let it leek please */
 	specs        = g_object_class_list_properties (object_class, &n_specs);
 
 	for (i = 0; i < n_specs; i++)
@@ -915,8 +899,8 @@ glade_widget_class_merge (GladeWidgetClass *widget_class,
 	if (widget_class->get_internal_child == NULL)
 		widget_class->get_internal_child = parent_class->get_internal_child;
 
-	if (widget_class->child_property_applies == NULL)
-		widget_class->child_property_applies = parent_class->child_property_applies;
+	if (widget_class->launch_editor == NULL)
+		widget_class->launch_editor = parent_class->launch_editor;
 
 	/* merge the parent's properties */
 	glade_widget_class_merge_properties
@@ -977,21 +961,14 @@ glade_widget_class_load_library (const gchar *library_name)
 	return module;
 }
 
-static gboolean
-glade_widget_class_direct_children (GtkWidget *ancestor, GtkWidget *widget, const char *property_id)
-{
-	if (ancestor != NULL && widget->parent == ancestor)
-		return TRUE;
-
-	return FALSE;
-}
-
 GladeWidgetClass *
-glade_widget_class_new (GladeXmlNode *class_node, const gchar *library)
+glade_widget_class_new (GladeXmlNode *class_node,
+			const gchar  *catname, 
+			const gchar  *library, 
+			const gchar  *domain)
 {
 	GladeWidgetClass *widget_class;
-	gchar            *name;
-	gchar            *generic_name;
+	gchar            *name, *generic_name, *ptr;
 	gchar            *title;
 	GModule          *module;
 	GType             parent_type;
@@ -1023,6 +1000,14 @@ glade_widget_class_new (GladeXmlNode *class_node, const gchar *library)
 						      GLADE_TAG_GENERIC_NAME);
 	title = glade_xml_get_property_string (class_node, GLADE_TAG_TITLE);
 
+	/* Translate title */
+	if (title != dgettext (domain, title))
+	{
+		ptr   = dgettext (domain, title);
+		g_free (title);
+		title = ptr;
+	}
+
 	module = NULL;
 	if (library) 
 	{
@@ -1044,8 +1029,12 @@ glade_widget_class_new (GladeXmlNode *class_node, const gchar *library)
 	widget_class->generic_name = generic_name;
 	widget_class->palette_name = title;
 	widget_class->in_palette   = title ? TRUE : FALSE;
+	widget_class->type         = glade_util_get_type_from_name (name);
+
+	/* Dont mention gtk+ as a required lib in the generated glade file */
+	if (strcmp (catname, "gtk+"))
+		widget_class->catalog = g_strdup (catname);
 
-	widget_class->type = glade_util_get_type_from_name (name);
 	if (widget_class->type == 0)
 	{
 		glade_widget_class_free (widget_class);
@@ -1056,7 +1045,6 @@ glade_widget_class_new (GladeXmlNode *class_node, const gchar *library)
 	widget_class->signals    = glade_widget_class_list_signals (widget_class);
 	widget_class->children   = glade_widget_class_list_children (widget_class);
 	widget_class->icon       = glade_widget_class_create_icon (widget_class);
-	widget_class->child_property_applies = glade_widget_class_direct_children;
 
 	for (parent_type = g_type_parent (widget_class->type);
 	     parent_type != 0;
@@ -1074,7 +1062,7 @@ glade_widget_class_new (GladeXmlNode *class_node, const gchar *library)
 	}
 	
 	if (glade_xml_node_get_children (class_node))
-		glade_widget_class_extend_with_node (widget_class, class_node);
+		glade_widget_class_extend_with_node (widget_class, class_node, domain);
 	
 	/* store the GladeWidgetClass on the cache,
 	 * if it's the first time we store a widget class, then
@@ -1120,7 +1108,8 @@ glade_widget_class_get_type (GladeWidgetClass *widget)
  *          class.
  */
 GladePropertyClass *
-glade_widget_class_get_property_class (GladeWidgetClass *class, const gchar *name)
+glade_widget_class_get_property_class (GladeWidgetClass *class, 
+				       const gchar      *name)
 {
 	GList *list, *l;
 	GladePropertyClass *pclass;
@@ -1324,12 +1313,17 @@ glade_widget_class_container_get_children (GladeWidgetClass *class,
 	return children;
 }
 
+
 GList *
 glade_widget_class_container_get_all_children (GladeWidgetClass *class,
 					       GObject          *container)
 {
 	GList *list, *children = NULL;
 
+
+	if (class->get_internal_children)
+		children = class->get_internal_children (container);
+
 	for (list = class->children; list && list->data; list = list->next)
 	{
 		GladeSupportedChild *support = list->data;
@@ -1337,7 +1331,9 @@ glade_widget_class_container_get_all_children (GladeWidgetClass *class,
 			children = g_list_concat (children,
 						  support->get_all_children (container));
 	}
-	return children;
+
+	/* Remove duplicates */
+	return glade_util_purify_list (children);
 }
 
 void
@@ -1414,13 +1410,13 @@ glade_widget_class_container_replace_child (GladeWidgetClass *class,
 }
 
 gboolean
-glade_widget_class_contains_non_widgets (GladeWidgetClass *class)
+glade_widget_class_contains_extra (GladeWidgetClass *class)
 {
 	GList *list;
 	for (list = class->children; list && list->data; list = list->next)
 	{
 		GladeSupportedChild *support = list->data;
-		if (!g_type_is_a (support->type, GTK_TYPE_WIDGET))
+		if (support->type != GTK_TYPE_WIDGET)
 			return TRUE;
 	}
 	return FALSE;
