@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  * Authors:
  *   Joaquin Cuenca Abela <e98cuenc@yahoo.com>
@@ -92,6 +92,13 @@ enum
 static guint glade_widget_signals[LAST_SIGNAL] = {0};
 static GObjectClass *parent_class = NULL;
 
+/* Sometimes we need to use the project deep in the loading code,
+ * this is just a shortcut way to get the project.
+ */
+static GladeProject *loading_project = NULL;
+
+
+
 /**
  * glade_widget_get_type:
  *
@@ -258,6 +265,8 @@ glade_widget_init (GladeWidget *widget)
 	widget->object = NULL;
 	widget->properties = NULL;
 	widget->packing_properties = NULL;
+	widget->prop_refs = NULL;
+	widget->prop_refs_readonly = FALSE;
 	widget->signals = g_hash_table_new_full
 		(g_str_hash, g_str_equal,
 		 (GDestroyNotify) g_free,
@@ -308,7 +317,6 @@ glade_widget_debug (GladeWidget *widget)
 	glade_widget_debug_real (widget, 0);
 }
 
-
 static void
 glade_widget_save_coords (GladeWidget *widget)
 {
@@ -355,6 +363,48 @@ glade_widget_hide (GladeWidget *widget)
 	widget->visible = FALSE;
 }
 
+
+void
+glade_widget_add_prop_ref (GladeWidget *widget, GladeProperty *property)
+{
+	g_return_if_fail (GLADE_IS_WIDGET (widget));
+	g_return_if_fail (GLADE_IS_PROPERTY (property));
+
+	if (property && !widget->prop_refs_readonly &&
+	    !g_list_find (widget->prop_refs, property))
+		widget->prop_refs = g_list_prepend (widget->prop_refs, property);
+}
+
+void
+glade_widget_remove_prop_ref (GladeWidget *widget, GladeProperty *property)
+{
+	g_return_if_fail (GLADE_IS_WIDGET (widget));
+	g_return_if_fail (GLADE_IS_PROPERTY (property));
+
+	if (!widget->prop_refs_readonly)
+		widget->prop_refs = g_list_remove (widget->prop_refs, property);
+}
+
+void
+glade_widget_project_notify (GladeWidget *widget, GladeProject *project)
+{
+	GList         *l;
+	GladeProperty *property;
+
+	widget->prop_refs_readonly = TRUE;
+	for (l = widget->prop_refs; l && l->data; l = l->next)
+	{
+		property = GLADE_PROPERTY (l->data);
+
+		if (project != NULL && 
+		    project == property->widget->project)
+			glade_property_set (property, widget->object);
+		else
+			glade_property_set (property, NULL);
+	}
+	widget->prop_refs_readonly = FALSE;
+}
+
 static void
 glade_widget_copy_packing_props (GladeWidget *parent,
 				 GladeWidget *child,
@@ -362,6 +412,8 @@ glade_widget_copy_packing_props (GladeWidget *parent,
 {
 	GList *l;
 
+	g_return_if_fail (child->parent == parent);
+
 	glade_widget_set_packing_properties (child, parent);
 
 	for (l = child->packing_properties; l && l->data; l = l->next)
@@ -395,17 +447,34 @@ glade_widget_sync_packing_props (GladeWidget *widget)
 	}
 }
 
-static void
-glade_widget_copy_props (GladeWidget *widget,
-			 GladeWidget *template)
+/**
+ * glade_widget_copy_properties:
+ * @widget:   a 'dest' #GladeWidget
+ * @template: a 'src' #GladeWidget
+ *
+ * Sets properties in @widget based on the values of
+ * matching properties in @template
+ */
+void
+glade_widget_copy_properties (GladeWidget *widget,
+			      GladeWidget *template)
 {
 	GList *l;
 	for (l = widget->properties; l && l->data; l = l->next)
 	{
-		GladeProperty *dup_prop  = GLADE_PROPERTY(l->data);
-		GladeProperty *orig_prop =
-			glade_widget_get_property (template, dup_prop->class->id);
-		glade_property_set_value (dup_prop, orig_prop->value);
+		GladeProperty *widget_prop  = GLADE_PROPERTY(l->data);
+		GladeProperty *template_prop;
+	
+		/* Check if they share the same class definition, different
+		 * properties may have the same name (support for
+		 * copying properties across "not-quite" compatible widget
+		 * classes, like GtkImageMenuItem --> GtkCheckMenuItem).
+		 */
+		if ((template_prop =
+		     glade_widget_get_property (template, 
+						widget_prop->class->id)) != NULL &&
+		    glade_property_class_match (template_prop->class, widget_prop->class))
+			glade_property_set_value (widget_prop, template_prop->value);
 	}
 }
 
@@ -490,7 +559,13 @@ glade_widget_build_object (GladeWidgetClass *klass, GladeWidget *widget)
 		{
 			if (g_value_type_compatible (G_VALUE_TYPE (glade_property_class->def),
 						     G_VALUE_TYPE (&parameter.value)))
+			{
+				if (g_type_is_a (G_VALUE_TYPE (glade_property_class->def), G_TYPE_OBJECT))
+					if (g_value_get_object (glade_property_class->def) == NULL)
+						continue;
+				
 				g_value_copy (glade_property_class->def, &parameter.value);
+			}
 			else
 			{
 				g_critical ("Type mismatch on %s property of %s",
@@ -552,7 +627,8 @@ glade_widget_set_default_packing_properties (GladeWidget *container,
 
 			/* Check value type */
 			value = glade_property_class_make_gvalue_from_string (property_class,
-									      def->value);
+									      def->value,
+									      child->project);
 
 			glade_widget_class_container_set_property (container->widget_class,
 								   container->object,
@@ -745,18 +821,24 @@ glade_widget_dup_internal (GladeWidget *parent, GladeWidget *template)
 									  child_dup->object);
 					
 				}
-				
-				/* Copy packing props of internal children as well ? XXX */
-				glade_widget_copy_packing_props (gwidget,
-								 child_dup,
-								 child_gwidget);
+
+				/* Internal children that are not heirarchic children
+				 * need to avoid copying these packing props (like popup windows
+				 * created on behalf of composite widgets).
+				 */
+				if (glade_widget_class_container_has_child (gwidget->widget_class,
+									    gwidget->object,
+									    child_dup->object))
+					glade_widget_copy_packing_props (gwidget,
+									 child_dup,
+									 child_gwidget);
 			}
 		}
 		g_list_free (children);
 	}
 
 	if (gwidget->internal)
-		glade_widget_copy_props (gwidget, template);
+		glade_widget_copy_properties (gwidget, template);
 	
 	/* If custom properties are still at thier
 	 * default value, they need to be synced.
@@ -1345,14 +1427,11 @@ glade_widget_get_pack_property (GladeWidget *widget, const gchar *id_property)
 	g_return_val_if_fail (GLADE_IS_WIDGET (widget), NULL);
 	g_return_val_if_fail (id_property != NULL, NULL);
 
-
 	for (list = widget->packing_properties; list; list = list->next) {
 		property = list->data;
 		if (strcmp (property->class->id, id_property) == 0)
 			return property;
 	}
-	g_critical ("Unable to find property %s on widget %s of class %s",
-		    id_property, widget->name, widget->widget_class->name);
 	return NULL;
 }
 
@@ -1384,6 +1463,8 @@ glade_widget_property_get (GladeWidget      *widget,
 		va_end (vl);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1414,6 +1495,8 @@ glade_widget_property_set (GladeWidget      *widget,
 		va_end (vl);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1444,6 +1527,8 @@ glade_widget_pack_property_get (GladeWidget      *widget,
 		va_end (vl);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1474,6 +1559,8 @@ glade_widget_pack_property_set (GladeWidget      *widget,
 		va_end (vl);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1504,6 +1591,8 @@ glade_widget_property_set_sensitive (GladeWidget      *widget,
 		glade_property_set_sensitive (property, sensitive, reason);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 
 }
@@ -1535,6 +1624,8 @@ glade_widget_pack_property_set_sensitive (GladeWidget      *widget,
 		glade_property_set_sensitive (property, sensitive, reason);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1564,6 +1655,8 @@ glade_widget_property_set_enabled (GladeWidget      *widget,
 		glade_property_set_enabled (property, enabled);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 
 }
@@ -1594,6 +1687,8 @@ glade_widget_pack_property_set_enabled (GladeWidget      *widget,
 		glade_property_set_enabled (property, enabled);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1620,6 +1715,8 @@ glade_widget_property_reset (GladeWidget   *widget,
 		glade_property_reset (property);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1645,6 +1742,8 @@ glade_widget_pack_property_reset (GladeWidget   *widget,
 		glade_property_reset (property);
 		return TRUE;
 	}
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1666,6 +1765,9 @@ glade_widget_property_default (GladeWidget *widget,
 
 	if ((property = glade_widget_get_property (widget, id_property)) != NULL)
 		return glade_property_default (property);
+
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1687,6 +1789,9 @@ glade_widget_pack_property_default (GladeWidget *widget,
 
 	if ((property = glade_widget_get_pack_property (widget, id_property)) != NULL)
 		return glade_property_default (property);
+
+	g_critical ("Unable to find property %s on widget %s of class %s",
+		    id_property, widget->name, widget->widget_class->name);
 	return FALSE;
 }
 
@@ -1789,6 +1894,9 @@ glade_widget_button_press (GtkWidget *widget,
 	gboolean           handled = FALSE;
 
 	glade_widget = glade_widget_retrieve_from_position (widget, x, y);
+
+	if (glade_widget == NULL) return FALSE;
+
 	widget = GTK_WIDGET (glade_widget_get_object (glade_widget));
 
 	/* make sure to grab focus, since we may stop default handlers */
@@ -1903,9 +2011,11 @@ glade_widget_connect_signal_handlers (GtkWidget *widget_gtk, gpointer data)
 /**
  * glade_widget_set_object:
  * @gwidget:
- * @new_widget:
+ * @new_object:
  *
- * TODO: write me
+ * Set the runtime object for this GladeWidget wrapper
+ * (this is used deep in the core and is probably unsafe
+ * to use elsewhere).
  */
 void
 glade_widget_set_object (GladeWidget *gwidget, GObject *new_object)
@@ -2174,6 +2284,36 @@ glade_widget_set_packing_properties (GladeWidget *widget,
 	}
 }
 
+/**
+ * glade_widget_has_decendant:
+ * @widget: a #GladeWidget
+ * @type: a  #GType
+ * 
+ * Returns: whether this GladeWidget has any decendants of type `type'
+ */
+gboolean
+glade_widget_has_decendant (GladeWidget *widget, GType type)
+{
+	GladeWidget   *child;
+	GList         *children, *l;
+	gboolean       found = FALSE;
+
+	if (g_type_is_a (widget->widget_class->type, type))
+		return TRUE;
+
+	if ((children = glade_widget_class_container_get_all_children
+	     (widget->widget_class, widget->object)) != NULL)
+	{
+		for (l = children; l; l = l->next)
+			if ((child = glade_widget_get_from_gobject (l->data)) != NULL &&
+			    (found = glade_widget_has_decendant (child, type)))
+				break;
+		g_list_free (children);
+	}	
+	return found;
+}
+
+
 /**
  * glade_widget_replace:
  * @old_object: a #GObject
@@ -2426,7 +2566,8 @@ glade_widget_value_from_prop_info (GladePropInfo    *info,
 	if (info->name && info->value &&
 	    (pclass = glade_widget_class_get_property_class (class, id)) != NULL)
 		gvalue = glade_property_class_make_gvalue_from_string (pclass,
-								       info->value);
+								       info->value,
+								       loading_project);
 		
 	g_free (id);
 
@@ -2462,6 +2603,15 @@ glade_widget_apply_property_from_prop_info (GladePropInfo *info,
 		widget->parent->widget_class :
 		widget->widget_class;
 	
+
+	if (glade_property_class_is_object (property->class))
+	{
+		/* we must synchronize this directly after loading this project.
+		 */
+		g_object_set_data_full (G_OBJECT (property), "glade-loaded-object", 
+					g_strdup (info->value), g_free);
+	}
+
 	if ((gvalue = glade_widget_value_from_prop_info (info, wclass)) != NULL)
 	{
 		glade_property_set_value (property, gvalue);
@@ -2552,16 +2702,12 @@ glade_widget_get_property_from_widget_info (GladeWidgetClass  *class,
 		{
 			g_free (id);
 
-			/* FIXME: Waiting for a solution in libglade for i18n
-			 * metadata.
-			 */
-			
 			if (translatable)
-				*translatable = FALSE; /*pinfo->translatable;*/
+				*translatable = pinfo->translatable;
 			if (has_context)
-				*has_context = FALSE; /*pinfo->has_context;*/
+				*has_context = pinfo->has_context;
 			if (comment)
-				*comment = NULL; /*g_strdup (pinfo->comment);*/
+				*comment = g_strdup (pinfo->comment);
 			
 			return glade_widget_value_from_prop_info (pinfo, class);
 		}
@@ -2596,20 +2742,27 @@ glade_widget_properties_from_widget_info (GladeWidgetClass *class,
 		gboolean            translatable;
 		gboolean            has_context;
 		gchar              *comment = NULL;
+		const gchar        *obj_name;
 
 		/* If there is a value in the XML, initialize property with it,
 		 * otherwise initialize property to default.
 		 */
-		value = glade_widget_get_property_from_widget_info (class,
-								    info,
-								    pclass->id,
-								    &translatable,
-								    &has_context,
-								    &comment);
+		value = glade_widget_get_property_from_widget_info 
+			(class, info, pclass->id,
+			 &translatable, &has_context, &comment);
 		
 		property          = glade_property_new (pclass, NULL, value);
 		property->enabled = value ? TRUE : property->enabled;
 
+		if (glade_property_class_is_object (pclass))
+		{
+			/* we must synchronize this directly after loading this project.
+			 */
+			obj_name = glade_parser_pvalue_from_winfo (info, pclass->id);
+			g_object_set_data_full (G_OBJECT (property), "glade-loaded-object", 
+						g_strdup (obj_name), g_free);
+		}
+
 		if (value) {
 			glade_property_i18n_set_translatable (property, translatable);
 			glade_property_i18n_set_has_context (property, has_context);
@@ -2646,7 +2799,9 @@ glade_widget_params_from_widget_info (GladeWidgetClass *widget_class,
 		glade_property_class =
 		     glade_widget_class_get_property_class (widget_class,
 							    pspec[i]->name);
-		if (!glade_property_class || glade_property_class->set_function)
+		if (glade_property_class == NULL ||
+		    glade_property_class->set_function ||
+		    glade_property_class->ignore)
 			continue;
 		
 		parameter.name = pspec[i]->name;
@@ -2841,8 +2996,10 @@ glade_widget_new_child_from_child_info (GladeChildInfo *info,
 				   info->internal_child, glade_widget_get_name (parent));
 			return FALSE;
 		}
-		child = glade_widget_new_for_internal_child 
-			(parent, child_object, info->internal_child);
+
+		if ((child = glade_widget_get_from_gobject (child_object)) == NULL)
+			g_error ("Unable to get GladeWidget for internal child %s\n",
+				 info->internal_child);
 
 		glade_widget_fill_from_widget_info (info->child, child, TRUE);
 		glade_widget_sync_custom_props (child);
@@ -2894,6 +3051,8 @@ GladeWidget *
 glade_widget_read (GladeProject *project, GladeWidgetInfo *info)
 {
 	GladeWidget *widget;
+
+	loading_project = project;
 		
 	if ((widget = glade_widget_new_from_widget_info
 	     (info, project, NULL)) != NULL)
@@ -2901,5 +3060,53 @@ glade_widget_read (GladeProject *project, GladeWidgetInfo *info)
 		if (glade_verbose)
 			glade_widget_debug (widget);
 	}	
+
+	loading_project = NULL;
+
 	return widget;
 }
+
+/**
+ * glade_widget_get_launcher:
+ * @widget: a #GladeWidget
+ *
+ * Returns: whether there is a #GladeEditorLaunchFunc provided by
+ *          the backend for @widget (or any of its parents)
+ */
+gboolean
+glade_widget_has_launcher (GladeWidget *widget)
+{
+	GladeWidget *parent;
+
+	g_return_val_if_fail (GLADE_IS_WIDGET (widget), FALSE);
+	parent = widget;
+	do
+	{
+		if (parent->widget_class->launch_editor != NULL)
+			return TRUE;
+	} while ((parent = parent->parent) != NULL);
+	return FALSE;
+}
+
+/**
+ * glade_widget_launch_editor:
+ * @widget: a #GladeWidget
+ *
+ * Launches a custom editor from the backend for thie widget.
+ */
+void
+glade_widget_launch_editor (GladeWidget *widget)
+{
+	GladeWidget *parent;
+
+	g_return_if_fail (GLADE_IS_WIDGET (widget));
+	parent = widget;
+	do
+	{
+		if (parent->widget_class->launch_editor != NULL)
+		{
+			parent->widget_class->launch_editor (parent->object);
+			break;
+		}
+	} while ((parent = parent->parent) != NULL);
+}
